import json
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    Iterable,
)
from pydantic import ValidationError
from requests.models import Response

from darwin.dataset import RemoteDataset
from darwin.dataset.release import Release
from darwin.dataset.upload_manager import (
    FileUploadCallback,
    ItemMergeMode,
    LocalFile,
    MultiFileItem,
    ProgressCallback,
    UploadHandler,
    UploadHandlerV2,
)
from darwin.dataset.utils import (
    chunk_items,
    get_external_file_name,
    get_external_file_type,
    is_relative_to,
    parse_external_file_path,
)
from darwin.datatypes import (
    AnnotationFile,
    ItemId,
    ObjectStore,
    PathLike,
    StorageKeyDictModel,
    StorageKeyListModel,
)
from darwin.exceptions import NotFound, UnknownExportVersion
from darwin.exporter.formats.darwin import build_image_annotation
from darwin.item import DatasetItem
from darwin.item_sorter import ItemSorter
from darwin.utils import (
    SUPPORTED_EXTENSIONS,
    PRESERVE_FOLDERS_KEY,
    AS_FRAMES_KEY,
    EXTRACT_VIEWS_KEY,
    find_files,
    urljoin,
)

if TYPE_CHECKING:
    from darwin.client import Client


class RemoteDatasetV2(RemoteDataset):
    """
    Manages the remote and local versions of a dataset hosted on Darwin.
    It allows several dataset management operations such as syncing between
    remote and local, pulling a remote dataset, removing the local files, ...

    Parameters
    ----------
    client : Client
        Client to use for interaction with the server.
    team : str
        Team the dataset belongs to.
    name : str
        Name of the datasets as originally displayed on Darwin.
        It may contain white spaces, capital letters and special characters, e.g. `Bird Species!`.
    slug : str
        This is the dataset name with everything lower-case, removed specials characters and
        spaces are replaced by dashes, e.g., `bird-species`. This string is unique within a team.
    dataset_id : int
        Unique internal reference from the Darwin backend.
    item_count : int, default: 0
        Dataset size (number of items).
    progress : float, default: 0
        How much of the dataset has been annotated 0.0 to 1.0 (1.0 == 100%).

    Attributes
    ----------
    client : Client
        Client to use for interaction with the server.
    team : str
        Team the dataset belongs to.
    name : str
        Name of the datasets as originally displayed on Darwin.
        It may contain white spaces, capital letters and special characters, e.g. `Bird Species!`.
    slug : str
        This is the dataset name with everything lower-case, removed specials characters and
        spaces are replaced by dashes, e.g., `bird-species`. This string is unique within a team.
    dataset_id : int
        Unique internal reference from the Darwin backend.
    item_count : int, default: 0
        Dataset size (number of items).
    progress : float, default: 0
        How much of the dataset has been annotated 0.0 to 1.0 (1.0 == 100%).
    """

    def __init__(
        self,
        *,
        client: "Client",
        team: str,
        name: str,
        slug: str,
        dataset_id: int,
        item_count: int = 0,
        progress: float = 0,
    ):
        super().__init__(
            client=client,
            team=team,
            name=name,
            slug=slug,
            dataset_id=dataset_id,
            item_count=item_count,
            progress=progress,
            version=2,
        )

    def get_releases(self, include_unavailable: bool = False) -> List["Release"]:
        """
        Get a sorted list of releases with the most recent first.

        Parameters
        ----------
        include_unavailable : bool, default: False
            If True, return  all releases, including those that are not available.

        Returns
        -------
        List["Release"]
            Returns a sorted list of available ``Release``\\s with the most recent first.
        """
        try:
            releases_json: List[Dict[str, Any]] = self.client.api_v2.get_exports(
                self.slug, team_slug=self.team
            )
        except NotFound:
            return []

        releases = [
            Release.parse_json(self.slug, self.team, payload)
            for payload in releases_json
        ]

        return sorted(
            (
                releases
                if include_unavailable
                else filter(lambda x: x.available, releases)
            ),
            key=lambda x: x.version,
            reverse=True,
        )

    def push(
        self,
        files_to_upload: Optional[Sequence[Union[PathLike, LocalFile]]],
        *,
        blocking: bool = True,
        multi_threaded: bool = True,
        max_workers: Optional[int] = None,
        fps: int = 0,
        as_frames: bool = False,
        extract_views: bool = False,
        handle_as_slices: Optional[bool] = False,
        files_to_exclude: Optional[List[PathLike]] = None,
        path: Optional[str] = None,
        preserve_folders: bool = False,
        progress_callback: Optional[ProgressCallback] = None,
        file_upload_callback: Optional[FileUploadCallback] = None,
        item_merge_mode: Optional[str] = None,
    ) -> UploadHandler:
        """
        Uploads a local dataset (images ONLY) in the datasets directory.

        Parameters
        ----------
        files_to_upload : Optional[List[Union[PathLike, LocalFile]]]
            List of files to upload. These can be folders.
            If `item_merge_mode` is set, these paths must be folders.
        blocking : bool, default: True
            If False, the dataset is not uploaded and a generator function is returned instead.
        multi_threaded : bool, default: True
            Uses multiprocessing to upload the dataset in parallel.
            If blocking is False this has no effect.
        max_workers : int, default: None
            Maximum number of workers to use for parallel upload.
        fps : int, default: 0
            When the uploading file is a video, specify its framerate.
        as_frames: bool, default: False
            When the uploading file is a video, specify whether it's going to be uploaded as a list of frames.
        extract_views: bool, default: False
            When the uploading file is a volume, specify whether it's going to be split into orthogonal views.
        handle_as_slices: Optioonal[bool], default: False
            Whether to upload DICOM files as slices
        files_to_exclude : Optional[PathLike]], default: None
            Optional list of files to exclude from the file scan. These can be folders.
        path: Optional[str], default: None
            Optional path to store the files in.
        preserve_folders : bool, default: False
            Specify whether or not to preserve folder paths when uploading
        progress_callback: Optional[ProgressCallback], default: None
            Optional callback, called every time the progress of an uploading files is reported.
        file_upload_callback: Optional[FileUploadCallback], default: None
            Optional callback, called every time a file chunk is uploaded.
        item_merge_mode : Optional[str]
            If set, each file path passed to `files_to_upload` behaves as follows:
            - Paths pointing directly to individual files are ignored
            - Paths pointing to folders of files will be uploaded according to the following mode rules.
            Note that folders will not be recursively searched, so only files in the first level of the folder will be uploaded:
                - "slots": Each file in the folder will be uploaded to a different slot of the same item.
                - "series": All `.dcm` files in the folder will be concatenated into a single slot. All other files are ignored.
                - "channels": Each file in the folder will be uploaded to a different channel of the same item.
        Returns
        -------
        handler : UploadHandler
            Class for handling uploads, progress and error messages.

        Raises
        ------
        ValueError
            - If ``files_to_upload`` is ``None``.
            - If a path is specified when uploading a LocalFile object.
            - If there are no files to upload (because path is wrong or the exclude filter excludes everything).
        """
        merge_incompatible_args = {
            PRESERVE_FOLDERS_KEY: preserve_folders,
            AS_FRAMES_KEY: as_frames,
            EXTRACT_VIEWS_KEY: extract_views,
        }

        if files_to_exclude is None:
            files_to_exclude = []

        if files_to_upload is None:
            raise ValueError("No files or directory specified.")

        if item_merge_mode:
            try:
                ItemMergeMode(item_merge_mode)
            except ValueError:
                raise ValueError(
                    f"Invalid item merge mode: {item_merge_mode}. Valid options are: 'slots', 'series', 'channels'"
                )
            incompatible_args = [
                arg for arg, value in merge_incompatible_args.items() if value
            ]

            if incompatible_args:
                incompatible_args_str = ", ".join(incompatible_args)
                raise TypeError(
                    f"`item_merge_mode` does not support the following incompatible arguments: {incompatible_args_str}."
                )

        # Folder paths
        search_files = [
            item for item in files_to_upload if not isinstance(item, LocalFile)
        ]

        if item_merge_mode:
            local_files, multi_file_items = _find_files_to_upload_as_multi_file_items(
                search_files, files_to_exclude, fps, item_merge_mode
            )
            handler = UploadHandlerV2(
                self, local_files, multi_file_items, handle_as_slices=handle_as_slices
            )
        else:
            local_files = _find_files_to_upload_as_single_file_items(
                search_files,
                files_to_upload,
                files_to_exclude,
                path,
                fps,
                as_frames,
                extract_views,
                preserve_folders,
            )
            handler = UploadHandlerV2(
                self, local_files, handle_as_slices=handle_as_slices
            )
        if blocking:
            handler.upload(
                max_workers=max_workers,
                multi_threaded=multi_threaded,
                progress_callback=progress_callback,
                file_upload_callback=file_upload_callback,
            )
        else:
            handler.prepare_upload()

        return handler

    def fetch_remote_files(
        self,
        filters: Optional[Dict[str, Union[str, List[str]]]] = None,
        sort: Optional[Union[str, ItemSorter]] = None,
    ) -> Iterator[DatasetItem]:
        """
        Fetch and lists all files on the remote dataset.

        Parameters
        ----------
        filters : Optional[Dict[str, Union[str, List[str]]]], default: None
            The filters to use. Files excluded by the filter won't be fetched.
        sort : Optional[Union[str, ItemSorter]], default: None
            A sorting direction. It can be a string with the values 'asc', 'ascending', 'desc',
            'descending' or an ``ItemSorter`` instance.

        Yields
        -------
        Iterator[DatasetItem]
            An iterator of ``DatasetItem``.
        """
        post_filters: List[Tuple[str, Any]] = []
        post_sort: Dict[str, str] = {}

        if filters:
            # Backward compatibility with V1 filter parameter
            if "filenames" in filters:
                filters["item_names"] = filters["filenames"]
                del filters["filenames"]

            for list_type in [
                "item_names",
                "statuses",
                "item_ids",
                "slot_types",
                "item_paths",
            ]:
                if list_type in filters:
                    if isinstance(filters[list_type], list):
                        for value in filters[list_type]:
                            post_filters.append(("{}[]".format(list_type), value))
                    else:
                        post_filters.append((list_type, str(filters[list_type])))

        if sort:
            item_sorter = ItemSorter.parse(sort)
            post_sort[f"sort[{item_sorter.field}]"] = item_sorter.direction.value
        cursor = {"page[size]": 500, "include_workflow_data": "true"}
        while True:
            query = post_filters + list(post_sort.items()) + list(cursor.items())
            response = self.client.api_v2.fetch_items(
                self.dataset_id, query, team_slug=self.team
            )
            yield from [
                DatasetItem.parse(item, dataset_slug=self.slug)
                for item in response["items"]
            ]

            if response["page"]["next"]:
                cursor["page[from]"] = response["page"]["next"]
            else:
                return

    def archive(self, items: Iterable[DatasetItem]) -> None:
        """
        Archives (soft-deletion) the given ``DatasetItem``\\s belonging to this ``RemoteDataset``.

        Parameters
        ----------
        items : Iterable[DatasetItem]
            The ``DatasetItem``\\s to be archived.
        """
        payload: Dict[str, Any] = {
            "filters": {
                "item_ids": [item.id for item in items],
                "dataset_ids": [self.dataset_id],
            }
        }
        self.client.api_v2.archive_items(payload, team_slug=self.team)

    def restore_archived(self, items: Iterable[DatasetItem]) -> None:
        """
        Restores the archived ``DatasetItem``\\s that belong to this ``RemoteDataset``.

        Parameters
        ----------
        items : Iterable[DatasetItem]
            The ``DatasetItem``\\s to be restored.
        """
        payload: Dict[str, Any] = {
            "filters": {
                "item_ids": [item.id for item in items],
                "dataset_ids": [self.dataset_id],
            }
        }
        self.client.api_v2.restore_archived_items(payload, team_slug=self.team)

    def move_to_new(self, items: Iterable[DatasetItem]) -> None:
        """
        Changes the given ``DatasetItem``\\s status to ``new``.

        Parameters
        ----------
        items : Iterable[DatasetItem]
            The ``DatasetItem``\\s whose status will change.
        """

        (workflow_id, stages) = self._fetch_stages("dataset")
        if not stages:
            raise ValueError("Dataset's workflow is missing a dataset stage")

        self.client.api_v2.move_to_stage(
            {"item_ids": [item.id for item in items], "dataset_ids": [self.dataset_id]},
            stages[0]["id"],
            workflow_id,
            team_slug=self.team,
        )

    def complete(self, items: Iterable[DatasetItem]) -> None:
        """
        Completes the given ``DatasetItem``\\s.

        Parameters
        ----------
        items : Iterable[DatasetItem]
            The ``DatasetItem``\\s to be completed.
        """
        (workflow_id, stages) = self._fetch_stages("complete")
        if not stages:
            raise ValueError("Dataset's workflow is missing a complete stage")

        self.client.api_v2.move_to_stage(
            {"item_ids": [item.id for item in items], "dataset_ids": [self.dataset_id]},
            stages[0]["id"],
            workflow_id,
            team_slug=self.team,
        )

    def delete_items(self, items: Iterable[DatasetItem]) -> None:
        """
        Deletes the given ``DatasetItem``\\s.

        Parameters
        ----------
        items : Iterable[DatasetItem]
            The ``DatasetItem``\\s to be deleted.
        """
        self.client.api_v2.delete_items(
            {"dataset_ids": [self.dataset_id], "item_ids": [item.id for item in items]},
            team_slug=self.team,
        )

    def export(
        self,
        name: str,
        annotation_class_ids: Optional[List[str]] = None,
        include_url_token: bool = False,
        include_authorship: bool = False,
        version: Optional[str] = None,
    ) -> None:
        """
        Create a new release for this ``RemoteDataset``.

        Parameters
        ----------
        name : str
            Name of the release.
        annotation_class_ids : Optional[List[str]], default: None
            List of the classes to filter.
        include_url_token : bool, default: False
            Should the image url in the export include a token enabling access without team
            membership or not?
        include_authorship : bool, default: False
            If set, include annotator and reviewer metadata for each annotation.
        version : Optional[str], default: None, enum: ["1.0", "2.0"]
            When used for V2 dataset, allows to force generation of either Darwin JSON 1.0 (Legacy) or newer 2.0.
            Omit this option to get your team's default.
        """
        str_version = str(version)
        if str_version == "2.0":
            format = "darwin_json_2"
        elif str_version == "1.0":
            format = "json"
        elif version is None:
            format = None
        else:
            raise UnknownExportVersion(version)

        filters = (
            None
            if not annotation_class_ids
            else {"annotation_class_ids": list(map(int, annotation_class_ids))}
        )

        self.client.api_v2.export_dataset(
            format=format,
            name=name,
            include_authorship=include_authorship,
            include_token=include_url_token,
            annotation_class_ids=None,
            filters=filters,
            dataset_slug=self.slug,
            team_slug=self.team,
        )

    def get_report(self, granularity: str = "day") -> str:
        """
        Returns a String representation of a CSV report for this ``RemoteDataset``.

        Parameters
        ----------
        granularity : str, default: "day"
            The granularity of the report, can be 'day', 'week' or 'month'.

        Returns
        -------
        str
            A CSV report.
        """
        response: Response = self.client.get_report(
            self.dataset_id, granularity, self.team
        )
        return response.text

    def workview_url_for_item(self, item: DatasetItem) -> str:
        """
        Returns the darwin URL for the given ``DatasetItem``.

        Parameters
        ----------
        item : DatasetItem
            The ``DatasetItem`` for which we want the url.

        Returns
        -------
        str
            The url.
        """
        return urljoin(
            self.client.base_url, f"/workview?dataset={self.dataset_id}&item={item.id}"
        )

    def post_comment(
        self,
        item: DatasetItem,
        text: str,
        x: float,
        y: float,
        w: float,
        h: float,
        slot_name: Optional[str] = None,
    ):
        """
        Adds a comment to an item in this dataset,
        Tries to infer slot_name if left out.
        """
        if not slot_name:
            if len(item.slots) != 1:
                raise ValueError(
                    f"Unable to infer slot for '{item.id}', has multiple slots: {','.join(item.slots)}"
                )
            slot_name = item.slots[0]["slot_name"]

        self.client.api_v2.post_comment(
            item.id, text, x, y, w, h, slot_name, team_slug=self.team
        )

    def import_annotation(self, item_id: ItemId, payload: Dict[str, Any]) -> None:
        """
        Imports the annotation for the item with the given id.

        Parameters
        ----------
        item_id: ItemId
            Identifier of the Item that we are import the annotation to.
        payload: Dict[str, Any]
            A dictionary with the annotation to import. The default format is:
            `{"annotations": serialized_annotations, "overwrite": "false"}`
        """

        self.client.api_v2.import_annotation(
            item_id, payload=payload, team_slug=self.team
        )

    def _fetch_stages(self, stage_type):
        detailed_dataset = self.client.api_v2.get_dataset(self.dataset_id)
        workflow_ids = detailed_dataset["workflow_ids"]
        if len(workflow_ids) == 0:
            raise ValueError("Dataset is not part of a workflow")
        # currently we can only be part of one workflow
        workflow_id = workflow_ids[0]
        workflow = self.client.api_v2.get_workflow(workflow_id, team_slug=self.team)
        return (
            workflow_id,
            [stage for stage in workflow["stages"] if stage["type"] == stage_type],
        )

    def _build_image_annotation(
        self, annotation_file: AnnotationFile, team_name: str
    ) -> Dict[str, Any]:
        return build_image_annotation(annotation_file, team_name)

    def register(
        self,
        object_store: ObjectStore,
        storage_keys: Union[List[str], Dict[str, List[str]]],
        fps: Optional[Union[str, float]] = None,
        multi_planar_view: bool = False,
        preserve_folders: bool = False,
        multi_slotted: bool = False,
    ) -> Dict[str, List[str]]:
        """
        Register files from external storage in a Darwin dataset.

        Parameters
        ----------
        object_store : ObjectStore
            Object store to use for the registration.
        storage_keys : List[str] | Dict[str, List[str]]
            Either:
            - Single-slotted items: A list of storage keys
            - Multi-slotted items: A dictionary with keys as item names and values as lists of storage keys
        fps : Optional[str], default: None
            When the uploading file is a video, specify its framerate.
        multi_planar_view : bool, default: False
            Uses multiplanar view when uploading files.
        preserve_folders : bool, default: False
            Specify whether or not to preserve folder paths when uploading.
        multi_slotted : bool, default: False
            Specify whether the items are multi-slotted or not.

        Returns
        -------
        Dict[str, List[str]]
            A dictionary with the list of registered files.

        Raises
        ------
        ValueError
            If the type of ``storage_keys``:
            - Isn't List[str] when ``multi_slotted`` is False.
            - Isn't Dict[str, List[str]] when ``multi_slotted`` is True.
        """
        if multi_slotted:
            try:
                StorageKeyDictModel(storage_keys=storage_keys)  # type: ignore
            except ValidationError as e:
                print(
                    f"Error validating storage keys: {e}\n\nPlease make sure your storage keys are a list of strings"
                )
                raise e
            results = self.register_multi_slotted(
                object_store,
                storage_keys,  # type: ignore
                fps,
                multi_planar_view,
                preserve_folders,
            )
            return results
        else:
            try:
                StorageKeyListModel(storage_keys=storage_keys)  # type: ignore
            except ValidationError as e:
                print(
                    f"Error validating storage keys: {e}\n\nPlease make sure your storage keys are a dictionary with keys as item names and values as lists of storage keys"
                )
                raise e
            results = self.register_single_slotted(
                object_store,
                storage_keys,  # type: ignore
                fps,
                multi_planar_view,
                preserve_folders,
            )
            return results

    def register_single_slotted(
        self,
        object_store: ObjectStore,
        storage_keys: List[str],
        fps: Optional[Union[str, float]] = None,
        multi_planar_view: bool = False,
        preserve_folders: bool = False,
    ) -> Dict[str, List[str]]:
        """
        Register files in the dataset in a single slot.

        Parameters
        ----------
        object_store : ObjectStore
            Object store to use for the registration.
        storage_keys : List[str]
            List of storage keys to register.
        fps : Optional[str], default: None
            When the uploading file is a video, specify its framerate.
        multi_planar_view : bool, default: False
            Uses multiplanar view when uploading files.
        preserve_folders : bool, default: False
            Specify whether or not to preserve folder paths when uploading

        Returns
        -------
        Dict[str, List[str]]
            A dictionary with the list of registered files.

        Raises
        ------
        TypeError
            If the file type of any storage keyis not supported.
        """
        items = []
        for storage_key in storage_keys:
            file_type = get_external_file_type(storage_key)
            if not file_type:
                raise TypeError(
                    f"Unsupported file type for the following storage key: {storage_key}.\nPlease make sure your storage key ends with one of the supported extensions:\n{SUPPORTED_EXTENSIONS}"
                )
            item = {
                "path": parse_external_file_path(storage_key, preserve_folders),
                "type": file_type,
                "storage_key": storage_key,
                "name": (
                    storage_key.split("/")[-1] if "/" in storage_key else storage_key
                ),
            }
            if fps and file_type == "video":
                item["fps"] = fps
            if multi_planar_view and file_type == "dicom":
                item["extract_views"] = "true"
            items.append(item)

        # Do not register more than 10 items in a single request
        chunk_size = 10
        chunked_items = chunk_items(items, chunk_size)
        print(f"Registering {len(items)} items in chunks of {chunk_size} items...")
        results = {
            "registered": [],
            "blocked": [],
        }

        for chunk in chunked_items:
            payload = {
                "items": chunk,
                "dataset_slug": self.slug,
                "storage_slug": object_store.name,
            }
            print(f"Registering {len(chunk)} items...")
            response = self.client.api_v2.register_items(payload, team_slug=self.team)
            for item in json.loads(response.text)["items"]:
                item_info = f"Item {item['name']} registered with item ID {item['id']}"
                results["registered"].append(item_info)
            for item in json.loads(response.text)["blocked_items"]:
                item_info = f"Item {item['name']} was blocked for the reason: {item['slots'][0]['reason']}"
                results["blocked"].append(item_info)
        print(
            f"{len(results['registered'])} of {len(storage_keys)} items registered successfully"
        )
        if results["blocked"]:
            print("The following items were blocked:")
            for item in results["blocked"]:
                print(f"  - {item}")
        print(f"Reistration complete. Check your items in the dataset: {self.slug}")
        return results

    def register_multi_slotted(
        self,
        object_store: ObjectStore,
        storage_keys: Dict[str, List[str]],
        fps: Optional[Union[str, float]] = None,
        multi_planar_view: bool = False,
        preserve_folders: bool = False,
    ) -> Dict[str, List[str]]:
        """
        Register files in the dataset in multiple slots.

        Parameters
        ----------
        object_store : ObjectStore
            Object store to use for the registration.
        storage_keys : Dict[str, List[str]
            Storage keys to register. The keys are the item names and the values are lists of storage keys.
        fps : Optional[str], default: None
            When the uploading file is a video, specify its framerate.
        multi_planar_view : bool, default: False
            Uses multiplanar view when uploading files.
        preserve_folders : bool, default: False
            Specify whether or not to preserve folder paths when uploading

        Returns
        -------
        Dict[str, List[str]]
            A dictionary with the list of registered files.

        Raises
        ------
        TypeError
            If the file type of any storage key is not supported.
        """
        items = []
        for item in storage_keys:
            slots = []
            for storage_key in storage_keys[item]:
                file_name = get_external_file_name(storage_key)
                file_type = get_external_file_type(storage_key)
                if not file_type:
                    raise TypeError(
                        f"Unsupported file type for the following storage key: {storage_key}.\nPlease make sure your storage key ends with one of the supported extensions:\n{SUPPORTED_EXTENSIONS}"
                    )
                slot = {
                    "slot_name": file_name,
                    "type": file_type,
                    "storage_key": storage_key,
                    "file_name": file_name,
                }
                if fps and file_type == "video":
                    slot["fps"] = fps
                if multi_planar_view and file_type == "dicom":
                    slot["extract_views"] = "true"
                slots.append(slot)
            items.append(
                {
                    "slots": slots,
                    "name": item,
                    "path": parse_external_file_path(
                        storage_keys[item][0], preserve_folders
                    ),
                }
            )

        # Do not register more than 10 items in a single request
        chunk_size = 10
        chunked_items = chunk_items(items, chunk_size)
        print(f"Registering {len(items)} items in chunks of {chunk_size} items...")
        results = {
            "registered": [],
            "blocked": [],
        }

        for chunk in chunked_items:
            payload = {
                "items": chunk,
                "dataset_slug": self.slug,
                "storage_slug": object_store.name,
            }
            print(f"Registering {len(chunk)} items...")
            response = self.client.api_v2.register_items(payload, team_slug=self.team)
            for item in json.loads(response.text)["items"]:
                item_info = f"Item {item['name']} registered with item ID {item['id']}"
                results["registered"].append(item_info)
            for item in json.loads(response.text)["blocked_items"]:
                item_info = f"Item {item['name']} was blocked for the reason: {item['slots'][0]['reason']}"
                results["blocked"].append(item_info)
        print(
            f"{len(results['registered'])} of {len(storage_keys)} items registered successfully"
        )
        if results["blocked"]:
            print("The following items were blocked:")
            for item in results["blocked"]:
                print(f"  - {item}")
        print(f"Reistration complete. Check your items in the dataset: {self.slug}")
        return results


def _find_files_to_upload_as_multi_file_items(
    search_files: List[PathLike],
    files_to_exclude: List[PathLike],
    fps: int,
    item_merge_mode: str,
) -> Tuple[List[LocalFile], List[MultiFileItem]]:
    """
    Finds files to upload according to the `item_merge_mode`.
    Does not search each directory recursively, only considers files in the first level of each directory.

    Parameters
    ----------
    search_files : List[PathLike]
        List of directories to search for files.
    files_to_exclude : List[PathLike]
        List of files to exclude from the file scan.
    fps : int
        When uploading video files, specify the framerate
    item_merge_mode : str
        Mode to merge the files in the folders. Valid options are: 'slots', 'series', 'channels'.

    Returns
    -------
    List[LocalFile]
        List of `LocalFile` objects contained within each `MultiFileItem`
    List[MultiFileItem]
        List of `MultiFileItem` objects to be uploaded
    """
    multi_file_items, local_files = [], []
    for directory in search_files:
        files_in_directory = list(
            find_files(
                [directory],
                files_to_exclude=files_to_exclude,
                recursive=False,
                sort=True,
            )
        )
        if not files_in_directory:
            print(
                f"Warning: There are no files in the first level of {directory}, skipping directory"
            )
            continue
        multi_file_item = MultiFileItem(
            Path(directory), files_in_directory, ItemMergeMode(item_merge_mode), fps
        )
        multi_file_items.append(multi_file_item)
        local_files.extend(multi_file_item.files)

    if not multi_file_items:
        raise ValueError(
            "No valid folders to upload after searching the passed directories for files"
        )
    return local_files, multi_file_items


def _find_files_to_upload_as_single_file_items(
    search_files: List[PathLike],
    files_to_upload: Optional[Sequence[Union[PathLike, LocalFile]]],
    files_to_exclude: List[PathLike],
    path: Optional[str],
    fps: int,
    as_frames: bool,
    extract_views: bool,
    preserve_folders: bool,
) -> List[LocalFile]:
    """
    Finds files to upload as single-slotted dataset items. Recursively searches the passed directories for files.

    Parameters
    ----------
    search_files : List[PathLike]
        List of directories to search for files.

    files_to_exclude : Optional[List[PathLike]]
        List of files to exclude from the file scan.
    files_to_upload : Optional[List[Union[PathLike, LocalFile]]]
        List of files to upload. These can be folders.
    path : Optional[str]
        Path to store the files in.
    fps: int
        When uploading video files, specify the framerate.
    as_frames: bool
        When uploading video files, specify whether to upload as a list of frames.
    extract_views: bool
        When uploading volume files, specify whether to split into orthogonal views.
    preserve_folders: bool
        Specify whether or not to preserve folder paths when uploading.

    Returns
    -------
    List[LocalFile]
        List of files to upload.
    """
    # Direct file paths
    uploading_files = [item for item in files_to_upload if isinstance(item, LocalFile)]

    generic_parameters_specified = (
        path is not None or fps != 0 or as_frames is not False
    )

    if (
        any(isinstance(item, LocalFile) for item in uploading_files)
        and generic_parameters_specified
    ):
        raise ValueError("Cannot specify a path when uploading a LocalFile object.")

    for found_file in find_files(search_files, files_to_exclude=files_to_exclude):
        local_path = path
        if preserve_folders:
            source_files = [
                source_file
                for source_file in search_files
                if is_relative_to(found_file, source_file)
            ]
            if source_files:
                local_path = str(
                    found_file.relative_to(source_files[0]).parent.as_posix()
                )
                if local_path == ".":
                    local_path = "/"
        uploading_files.append(
            LocalFile(
                found_file,
                fps=fps,
                as_frames=as_frames,
                extract_views=extract_views,
                path=local_path,
            )
        )

    if not uploading_files:
        raise ValueError(
            "No files to upload, check your path, exclusion filters and resume flag"
        )

    return uploading_files



from darwin.future.core.properties.create import create_property
from darwin.future.core.properties.get import (
    get_property_by_id,
    get_team_full_properties,
    get_team_properties,
)
from darwin.future.core.properties.update import update_property, update_property_value



import responses

from darwin.future.core.client import ClientCore
from darwin.future.core.properties import create_property
from darwin.future.data_objects.properties import FullProperty
from darwin.future.tests.core.fixtures import *


@responses.activate
def test_create_property(
    base_client: ClientCore, base_property_object: FullProperty
) -> None:
    # Mocking the response using responses library
    responses.add(
        responses.POST,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/properties",
        json=base_property_object.model_dump(mode="json"),
        status=200,
    )
    # Call the function being tested
    property = create_property(
        base_client,
        params=base_property_object,
        team_slug=base_client.config.default_team,
    )

    # Assertions
    assert isinstance(property, FullProperty)
    assert property == base_property_object


@responses.activate
def test_create_property_from_json(
    base_client: ClientCore, base_property_object: FullProperty
) -> None:
    json = base_property_object.to_create_endpoint()
    # Mocking the response using responses library
    responses.add(
        responses.POST,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/properties",
        json=base_property_object.model_dump(mode="json"),
        status=200,
    )
    # Call the function being tested
    property = create_property(
        base_client, params=json, team_slug=base_client.config.default_team
    )

    # Assertions
    assert isinstance(property, FullProperty)
    assert property == base_property_object



# Can't import * in this module because of a circular import problem specific to teams
# The TeamCore module can instantiate from a client, but the client needs to use the
# team backend module to request the object for team. To circumvent this there's a
# get_raw method in this module that returns the raw team object, which is then passed
# to the TeamCore module, but if we import * here it introduces the
# circular import problem.



from uuid import UUID

from darwin.future.meta.client import Client
from darwin.future.meta.objects.v7_id import V7ID
from darwin.future.tests.meta.fixtures import *


def test_v7_id(base_meta_client: Client) -> None:
    # Test creating a V7ID object
    uuid = UUID("123e4567-e89b-12d3-a456-426655440000")
    v7_id = V7ID(uuid, base_meta_client)
    assert v7_id.id == uuid

    # Test __str__ method
    assert str(v7_id) == str(uuid)

    # Test __repr__ method
    assert repr(v7_id) == str(uuid)



from datetime import datetime, timedelta
from typing import List, Tuple, get_args

import pytest
from pydantic import ValidationError

from darwin.future.data_objects import advanced_filters as AF

FileTypesParameters: List[Tuple[AF.AcceptedFileTypes]] = list(
    get_args(AF.AcceptedFileTypes)
)

IssueTypes: List[Tuple[AF.IssueType]] = list(get_args(AF.IssueType))
ProcessingStatusTypes: List[Tuple[AF.ProcessingStatusType]] = list(
    get_args(AF.ProcessingStatusType)
)
WorkflowStatusTypes: List[Tuple[AF.WorkflowStatusType]] = list(
    get_args(AF.WorkflowStatusType)
)


def test_date_validator() -> None:
    with pytest.raises(ValidationError):
        AF.DateRange()  # needs at least one date

    with pytest.raises(ValidationError):
        # start date must be before end date
        start = datetime.now()
        end = start - timedelta(days=1)
        AF.DateRange(start=start, end=end)

    # Test instantiates
    AF.DateRange(start=datetime.now())
    AF.DateRange(end=datetime.now())
    AF.DateRange(start=datetime.now(), end=datetime.now())


def test_any_of_validator() -> None:
    with pytest.raises(ValidationError):
        AF.AnyOf[int](values=[])  # needs at least one value

    # Test instantiates
    AF.AnyOf[int](values=[1, 2])


def test_none_of_validator() -> None:
    with pytest.raises(ValidationError):
        AF.NoneOf[int](values=[])  # needs at least one value

    # Test instantiates
    AF.NoneOf[int](values=[1])


def test_all_of_validator() -> None:
    with pytest.raises(ValidationError):
        AF.AllOf[int](values=[])  # needs at least one value

    # Test instantiates
    AF.AllOf[int](values=[1])


def test_group_validator() -> None:
    with pytest.raises(ValidationError):
        AF.GroupFilter(filters=[])  # needs at least one filter

    # Test instantiates
    AF.GroupFilter(filters=[AF.AnnotationClass.all_of([1, 2]), AF.Assignee.any_of([1])])


def test_bitwise_syntax() -> None:
    # test 'and'
    sf = AF.AnnotationClass.all_of([1, 2, 3])
    gf_and = AF.GroupFilter(conjunction="and", filters=[sf, sf])
    assert gf_and == sf & sf
    assert gf_and & sf == sf & sf & sf
    assert gf_and | sf == AF.GroupFilter(conjunction="or", filters=[gf_and, sf])

    # test 'or'
    gf_or = AF.GroupFilter(conjunction="or", filters=[sf, sf])
    assert gf_or == sf | sf
    assert gf_or | sf == sf | sf | sf
    assert gf_or & sf == AF.GroupFilter(conjunction="and", filters=[gf_or, sf])

    # test mixed 'and' and 'or'
    assert gf_and & gf_or == AF.GroupFilter(conjunction="and", filters=[gf_and, gf_or])
    assert gf_or & gf_and == AF.GroupFilter(conjunction="and", filters=[gf_or, gf_and])
    assert gf_and | gf_or == AF.GroupFilter(conjunction="or", filters=[gf_and, gf_or])
    assert gf_or | gf_and == AF.GroupFilter(conjunction="or", filters=[gf_or, gf_and])


# test instantiatiors and factories
def test_annotation_class() -> None:
    assert AF.AnnotationClass.all_of([1, 2, 3]) == AF.AnnotationClass(
        matcher=AF.AllOf[int](values=[1, 2, 3])
    )
    assert AF.AnnotationClass.none_of([1, 2, 3]) == AF.AnnotationClass(
        matcher=AF.NoneOf[int](values=[1, 2, 3])
    )
    assert AF.AnnotationClass.any_of([1, 2, 3]) == AF.AnnotationClass(
        matcher=AF.AnyOf[int](values=[1, 2, 3])
    )


def test_assignee() -> None:
    assert AF.Assignee.all_of([1, 2, 3]) == AF.Assignee(
        matcher=AF.AllOf[int](values=[1, 2, 3])
    )
    assert AF.Assignee.none_of([1, 2, 3]) == AF.Assignee(
        matcher=AF.NoneOf[int](values=[1, 2, 3])
    )
    assert AF.Assignee.any_of([1, 2, 3]) == AF.Assignee(
        matcher=AF.AnyOf[int](values=[1, 2, 3])
    )


def test_created_at() -> None:
    after = datetime.now()
    before = after - timedelta(days=1)
    assert AF.CreatedAt.between(before, after) == AF.CreatedAt(
        matcher=AF.DateRange(start=before, end=after)
    )
    assert AF.CreatedAt.after(before) == AF.CreatedAt(
        matcher=AF.DateRange(start=before)
    )
    assert AF.CreatedAt.before(after) == AF.CreatedAt(matcher=AF.DateRange(end=after))

    # Check raises if time is not in order
    with pytest.raises(ValidationError):
        AF.CreatedAt.between(after, before)


def test_current_assignee() -> None:
    assert AF.CurrentAssignee.none_of([1, 2, 3]) == AF.CurrentAssignee(
        matcher=AF.NoneOf[int](values=[1, 2, 3])
    )
    assert AF.CurrentAssignee.any_of([1, 2, 3]) == AF.CurrentAssignee(
        matcher=AF.AnyOf[int](values=[1, 2, 3])
    )


@pytest.mark.parametrize("file_type", FileTypesParameters)
def test_file_type(file_type: AF.AcceptedFileTypes) -> None:
    assert AF.FileType.any_of([file_type]) == AF.FileType(
        matcher=AF.AnyOf[AF.AcceptedFileTypes](values=[file_type])
    )
    assert AF.FileType.none_of([file_type]) == AF.FileType(
        matcher=AF.NoneOf[AF.AcceptedFileTypes](values=[file_type])
    )
    assert AF.FileType.any_of([file_type]) == AF.FileType(
        matcher=AF.AnyOf[AF.AcceptedFileTypes](values=[file_type])
    )


def test_file_type_breaks() -> None:
    with pytest.raises(ValidationError):
        AF.FileType.any_of(["not a valid filetype"])  # type: ignore


def test_folder_path() -> None:
    assert AF.FolderPath.any_of(["test"]) == AF.FolderPath(
        matcher=AF.AnyOf[str](values=["test"])
    )
    assert AF.FolderPath.none_of(["test"]) == AF.FolderPath(
        matcher=AF.NoneOf[str](values=["test"])
    )
    assert AF.FolderPath.any_of(["test"]) == AF.FolderPath(
        matcher=AF.AnyOf[str](values=["test"])
    )


def test_id() -> None:
    assert AF.ID.any_of(["test"]) == AF.ID(matcher=AF.AnyOf[str](values=["test"]))
    assert AF.ID.none_of(["test"]) == AF.ID(matcher=AF.NoneOf[str](values=["test"]))


@pytest.mark.parametrize("issue_type", IssueTypes)
def test_issue(issue_type: AF.IssueType) -> None:
    assert AF.Issue.any_of([issue_type]) == AF.Issue(
        matcher=AF.AnyOf[AF.IssueType](values=[issue_type])
    )
    assert AF.Issue.none_of([issue_type]) == AF.Issue(
        matcher=AF.NoneOf[AF.IssueType](values=[issue_type])
    )


def test_issue_breaks() -> None:
    with pytest.raises(ValidationError):
        AF.Issue.any_of(["not a valid issue type"])  # type: ignore


def test_item_name() -> None:
    assert AF.ItemName.any_of(["test"]) == AF.ItemName(
        matcher=AF.AnyOf[str](values=["test"])
    )
    assert AF.ItemName.none_of(["test"]) == AF.ItemName(
        matcher=AF.NoneOf[str](values=["test"])
    )
    assert AF.ItemName.prefix("test") == AF.ItemName(matcher=AF.Prefix(value="test"))
    assert AF.ItemName.suffix("test") == AF.ItemName(matcher=AF.Suffix(value="test"))
    assert AF.ItemName.contains("test") == AF.ItemName(
        matcher=AF.Contains(value="test")
    )
    assert AF.ItemName.not_contains("test") == AF.ItemName(
        matcher=AF.NotContains(value="test")
    )


@pytest.mark.parametrize("processing_status", ProcessingStatusTypes)
def test_processing_status(processing_status: AF.ProcessingStatusType) -> None:
    assert AF.ProcessingStatus.any_of([processing_status]) == AF.ProcessingStatus(
        matcher=AF.AnyOf[AF.ProcessingStatusType](values=[processing_status])
    )
    assert AF.ProcessingStatus.none_of([processing_status]) == AF.ProcessingStatus(
        matcher=AF.NoneOf[AF.ProcessingStatusType](values=[processing_status])
    )


def test_processing_status_breaks() -> None:
    with pytest.raises(ValidationError):
        AF.ProcessingStatus.any_of(["not a valid processing status"])  # type: ignore


def test_updated_at() -> None:
    after = datetime.now()
    before = after - timedelta(days=1)
    assert AF.UpdatedAt.between(before, after) == AF.UpdatedAt(
        matcher=AF.DateRange(start=before, end=after)
    )
    assert AF.UpdatedAt.after(before) == AF.UpdatedAt(
        matcher=AF.DateRange(start=before)
    )
    assert AF.UpdatedAt.before(after) == AF.UpdatedAt(matcher=AF.DateRange(end=after))

    # Check raises if time is not in order
    with pytest.raises(ValidationError):
        AF.UpdatedAt.between(after, before)


@pytest.mark.parametrize("workflow_status", WorkflowStatusTypes)
def test_workflow_status(workflow_status: AF.WorkflowStatusType) -> None:
    assert AF.WorkflowStatus.any_of([workflow_status]) == AF.WorkflowStatus(
        matcher=AF.AnyOf[AF.WorkflowStatusType](values=[workflow_status])
    )
    assert AF.WorkflowStatus.none_of([workflow_status]) == AF.WorkflowStatus(
        matcher=AF.NoneOf[AF.WorkflowStatusType](values=[workflow_status])
    )


def test_workflow_status_breaks() -> None:
    with pytest.raises(ValidationError):
        AF.WorkflowStatus.any_of(["not a valid workflow status"])  # type: ignore


def test_workflow_stage() -> None:
    assert AF.WorkflowStage.any_of(["test"]) == AF.WorkflowStage(
        matcher=AF.AnyOf[str](values=["test"])
    )
    assert AF.WorkflowStage.none_of(["test"]) == AF.WorkflowStage(
        matcher=AF.NoneOf[str](values=["test"])
    )


def test_GF_validators() -> None:
    with pytest.raises(ValidationError):
        AF.GroupFilter(filters=[])  # needs at least two filters
        AF.GroupFilter(filters=[AF.AnnotationClass.all_of([1, 2])])



from __future__ import annotations

from typing import Dict, List
from uuid import UUID

from darwin.future.core.client import ClientCore
from darwin.future.core.types.common import JSONType
from darwin.future.data_objects.typing import UnknownType


def move_items_to_stage(
    client: ClientCore,
    team_slug: str,
    workflow_id: UUID,
    dataset_ids: int | List[int],
    stage_id: UUID,
    filters: Dict[str, UnknownType] = {},
) -> JSONType:
    """
    Moves a list of items to a stage

    Parameters
    ----------
    client: Client
        The client to use for the request.
    team_slug: str
        The slug of the team to move items for.
    workflow_id: UUID
        The id of the workflow to move items for.
    dataset_ids: int | List[int]
        The ID(s) of the dataset(s) containing the items.
    stage_id: UUID
        The id of the workflow to move items for.
    filters: Dict[str, UnknownType]
        Filter parameters.

    Returns
    -------
    JSONType
        The response data.
    """
    assert (
        filters
    ), "No parameters provided, please provide at least one non-dataset id filter"
    payload = {
        "filters": {
            "dataset_ids": (
                dataset_ids if isinstance(dataset_ids, list) else [dataset_ids]
            ),
            **filters,
        },
        "stage_id": str(stage_id),
        "workflow_id": str(workflow_id),
    }

    return client.post(f"/v2/teams/{team_slug}/items/stage", data=payload)



from __future__ import annotations

from typing import Any, Dict, List, Mapping, Protocol, Union

from darwin.future.data_objects import validators as darwin_validators

JSONType = Union[Dict[str, Any], List[Dict[str, Any]]]  # type: ignore
JSONDict = Dict[str, Any]  # type: ignore


class Implements_str(Protocol):
    def __str__(self) -> str: ...


Stringable = Union[str, Implements_str]


class TeamSlug(str):
    """
    Represents a team slug, which is a string identifier for a team.

    Attributes:
    -----------
    min_length : int
        The minimum length of a valid team slug.
    max_length : int
        The maximum length of a valid team slug.

    Methods:
    --------
    __get_validators__() -> generator
        Returns a generator that yields the validator function for this model.
    validate(v: str) -> TeamSlug
        Validates the input string and returns a new TeamSlug object.
    __repr__() -> str
        Returns a string representation of the TeamSlug object.
    """

    min_length = 1
    max_length = 256

    @classmethod
    def validate(cls, v: str) -> "TeamSlug":
        assert (
            len(v) < cls.max_length
        ), f"maximum length for team slug is {cls.max_length}"
        assert (
            len(v) > cls.min_length
        ), f"minimum length for team slug is {cls.min_length}"
        if not isinstance(v, str):
            raise TypeError("string required")
        modified_value = darwin_validators.parse_name(v)
        return cls(modified_value)

    def __repr__(self) -> str:
        return f"TeamSlug({super().__repr__()})"


class QueryString:
    """
    Represents a query string, which is a dictionary of string key-value pairs.

    Attributes:
    -----------
    value : Dict[str, str]
        The dictionary of key-value pairs that make up the query string.

    Methods:
    --------
    dict_check(value: Any) -> Dict[str, str]
        Validates that the input value is a dictionary of string key-value pairs.
        Returns the validated dictionary.
    __init__(value: Dict[str, str]) -> None
        Initializes a new QueryString object with the given dictionary of key-value pairs.
    __str__() -> str
        Returns a string representation of the QueryString object, in the format "?key1=value1&key2=value2".
    """

    value: dict[str, list[str] | str]

    def dict_check(
        self, value: Mapping[str, list[Stringable] | Stringable]
    ) -> dict[str, list[str] | str]:
        mapped: dict[str, list[str] | str] = {}
        for k, v in value.items():
            if isinstance(v, list):
                mapped[k] = [str(x) for x in v]
            else:
                mapped[k] = str(v)
        return mapped

    def __init__(self, value: Mapping[str, list[Stringable] | Stringable]) -> None:
        self.value = self.dict_check(value)

    def __str__(self) -> str:
        output: str = "?" if self.value else ""
        for k, v in self.value.items():
            if isinstance(v, list):
                for x in v:
                    output += f"{k}={x.lower()}&"
            else:
                output += f"{k}={v.lower()}&"
        return output[:-1]  # remove trailing &

    def __add__(self, other: QueryString) -> QueryString:
        return QueryString({**self.value, **other.value})

    def get(self, key: str, default: str = "") -> List[str] | str:
        return self.value.get(key, default)



from __future__ import annotations

from datetime import datetime
from typing import Generic, List, Literal, Optional, TypeVar, Union

from pydantic import BaseModel, SerializeAsAny, field_validator, model_validator

T = TypeVar("T")

AcceptedFileTypes = Literal["image", "video", "pdf", "dicom"]
IssueType = Literal["comment"]
ProcessingStatusType = Literal[
    "cancelled", "error", "uploading", "uploading_confirmed", "processing", "complete"
]
WorkflowStatusType = Literal["new", "annotate", "review", "complete"]


def validate_at_least_one(value: list[T]) -> list[T]:
    if len(value) < 1:
        raise ValueError("Must provide at least one value.")
    return list(set(value))


class BaseMatcher(BaseModel):
    name: str


class AnyOf(BaseMatcher, Generic[T]):
    name: Literal["any_of"] = "any_of"
    values: List[T]

    _normalize_values = field_validator("values")(validate_at_least_one)


class AllOf(BaseMatcher, Generic[T]):
    name: Literal["all_of"] = "all_of"
    values: List[T]

    _normalize_values = field_validator("values")(validate_at_least_one)


class NoneOf(BaseMatcher, Generic[T]):
    name: Literal["none_of"] = "none_of"
    values: List[T]

    _normalize_values = field_validator("values")(validate_at_least_one)


class Equals(BaseMatcher, Generic[T]):
    name: Literal["equals"] = "equals"
    value: T


class DateRange(BaseMatcher):
    name: Literal["date_range"] = "date_range"
    start: Optional[datetime] = None
    end: Optional[datetime] = None

    @model_validator(mode="before")
    def validate_date_range(cls, values: dict) -> dict:
        if not values.get("start") and not values.get("end"):
            raise ValueError("At least one of 'start' or 'end' must be provided.")
        if values.get("start") and values.get("end"):
            if values["start"] > values["end"]:
                raise ValueError("'start' must be before 'end'.")
        return values


class Prefix(BaseMatcher):
    name: Literal["prefix"] = "prefix"
    value: str


class Suffix(BaseMatcher):
    name: Literal["suffix"] = "suffix"
    value: str


class Contains(BaseMatcher):
    name: Literal["contains"] = "contains"
    value: str


class NotContains(BaseMatcher):
    name: Literal["not_contains"] = "not_contains"
    value: str


class SubjectFilter(BaseModel):
    subject: str
    matcher: SerializeAsAny[BaseMatcher]

    def __and__(self, other: SubjectFilter | GroupFilter) -> GroupFilter:
        if isinstance(other, GroupFilter):
            return GroupFilter(conjunction="and", filters=[self, other])
        return GroupFilter(conjunction="and", filters=[self, other])

    def __or__(self, other: SubjectFilter | GroupFilter) -> GroupFilter:
        if isinstance(other, GroupFilter):
            return GroupFilter(conjunction="or", filters=[self, other])
        return GroupFilter(conjunction="or", filters=[self, other])


# Subject Filters
class AnnotationClass(SubjectFilter):
    subject: Literal["annotation_class"] = "annotation_class"
    matcher: Union[AnyOf[int], AllOf[int], NoneOf[int]]

    @classmethod
    def any_of(cls, values: list[int]) -> AnnotationClass:
        return AnnotationClass(
            subject="annotation_class", matcher=AnyOf[int](values=values)
        )

    @classmethod
    def all_of(cls, values: list[int]) -> AnnotationClass:
        return AnnotationClass(
            subject="annotation_class", matcher=AllOf[int](values=values)
        )

    @classmethod
    def none_of(cls, values: list[int]) -> AnnotationClass:
        return AnnotationClass(
            subject="annotation_class", matcher=NoneOf[int](values=values)
        )


class Archived(SubjectFilter):
    subject: Literal["archived"] = "archived"
    matcher: Equals[bool]

    @classmethod
    def equals(cls, value: bool) -> Archived:
        return Archived(subject="archived", matcher=Equals(value=value))


class Assignee(SubjectFilter):
    subject: Literal["assignee"] = "assignee"
    matcher: Union[AnyOf[int], AllOf[int], NoneOf[int]]

    @classmethod
    def any_of(cls, values: list[int]) -> Assignee:
        return Assignee(subject="assignee", matcher=AnyOf[int](values=values))

    @classmethod
    def all_of(cls, values: list[int]) -> Assignee:
        return Assignee(subject="assignee", matcher=AllOf[int](values=values))

    @classmethod
    def none_of(cls, values: list[int]) -> Assignee:
        return Assignee(subject="assignee", matcher=NoneOf[int](values=values))


class CreatedAt(SubjectFilter):
    subject: Literal["created_at"] = "created_at"
    matcher: DateRange

    @classmethod
    def between(cls, start: datetime, end: datetime) -> CreatedAt:
        return CreatedAt(subject="created_at", matcher=DateRange(start=start, end=end))

    @classmethod
    def before(cls, end: datetime) -> CreatedAt:
        return CreatedAt(subject="created_at", matcher=DateRange(end=end))

    @classmethod
    def after(cls, start: datetime) -> CreatedAt:
        return CreatedAt(subject="created_at", matcher=DateRange(start=start))


class CurrentAssignee(SubjectFilter):
    subject: Literal["current_assignee"] = "current_assignee"
    matcher: Union[AnyOf[int], NoneOf[int]]

    @classmethod
    def any_of(cls, values: list[int]) -> CurrentAssignee:
        return CurrentAssignee(
            subject="current_assignee", matcher=AnyOf[int](values=values)
        )

    @classmethod
    def none_of(cls, values: list[int]) -> CurrentAssignee:
        return CurrentAssignee(
            subject="current_assignee", matcher=NoneOf[int](values=values)
        )


class FileType(SubjectFilter):
    subject: Literal["file_type"] = "file_type"
    matcher: Union[
        AnyOf[AcceptedFileTypes], AllOf[AcceptedFileTypes], NoneOf[AcceptedFileTypes]
    ]

    @classmethod
    def any_of(cls, values: list[AcceptedFileTypes]) -> FileType:
        return FileType(
            subject="file_type", matcher=AnyOf[AcceptedFileTypes](values=values)
        )

    @classmethod
    def all_of(cls, values: list[AcceptedFileTypes]) -> FileType:
        return FileType(
            subject="file_type", matcher=AllOf[AcceptedFileTypes](values=values)
        )

    @classmethod
    def none_of(cls, values: list[AcceptedFileTypes]) -> FileType:
        return FileType(
            subject="file_type", matcher=NoneOf[AcceptedFileTypes](values=values)
        )


class FolderPath(SubjectFilter):
    subject: Literal["folder_path"] = "folder_path"
    matcher: Union[AnyOf[str], NoneOf[str], Prefix, Suffix]

    @classmethod
    def any_of(cls, values: list[str]) -> FolderPath:
        return FolderPath(subject="folder_path", matcher=AnyOf[str](values=values))

    @classmethod
    def none_of(cls, values: list[str]) -> FolderPath:
        return FolderPath(subject="folder_path", matcher=NoneOf[str](values=values))

    @classmethod
    def prefix(cls, value: str) -> FolderPath:
        return FolderPath(subject="folder_path", matcher=Prefix[str](value=value))

    @classmethod
    def suffix(cls, value: str) -> FolderPath:
        return FolderPath(subject="folder_path", matcher=Suffix[str](value=value))


class ID(SubjectFilter):
    subject: Literal["id"] = "id"
    matcher: Union[AnyOf[str], NoneOf[str]]

    @classmethod
    def any_of(cls, values: list[str]) -> ID:
        return ID(subject="id", matcher=AnyOf[str](values=values))

    @classmethod
    def none_of(cls, values: list[str]) -> ID:
        return ID(subject="id", matcher=NoneOf[str](values=values))


class Issue(SubjectFilter):
    subject: Literal["issue"] = "issue"
    matcher: Union[AnyOf[IssueType], NoneOf[IssueType]]

    @classmethod
    def any_of(cls, values: list[IssueType]) -> Issue:
        return Issue(subject="issue", matcher=AnyOf[IssueType](values=values))

    @classmethod
    def none_of(cls, values: list[IssueType]) -> Issue:
        return Issue(subject="issue", matcher=NoneOf[IssueType](values=values))


class ItemName(SubjectFilter):
    subject: Literal["item_name"] = "item_name"
    matcher: Union[AnyOf[str], NoneOf[str], Prefix, Suffix, Contains, NotContains]

    @classmethod
    def any_of(cls, values: list[str]) -> ItemName:
        return ItemName(subject="item_name", matcher=AnyOf[str](values=values))

    @classmethod
    def none_of(cls, values: list[str]) -> ItemName:
        return ItemName(subject="item_name", matcher=NoneOf[str](values=values))

    @classmethod
    def prefix(cls, value: str) -> ItemName:
        return ItemName(subject="item_name", matcher=Prefix(value=value))

    @classmethod
    def suffix(cls, value: str) -> ItemName:
        return ItemName(subject="item_name", matcher=Suffix(value=value))

    @classmethod
    def contains(cls, value: str) -> ItemName:
        return ItemName(subject="item_name", matcher=Contains(value=value))

    @classmethod
    def not_contains(cls, value: str) -> ItemName:
        return ItemName(subject="item_name", matcher=NotContains(value=value))


class ProcessingStatus(SubjectFilter):
    subject: Literal["processing_status"] = "processing_status"
    matcher: Union[AnyOf[ProcessingStatusType], NoneOf[ProcessingStatusType]]

    @classmethod
    def any_of(cls, values: list[ProcessingStatusType]) -> ProcessingStatus:
        return ProcessingStatus(
            subject="processing_status",
            matcher=AnyOf[ProcessingStatusType](values=values),
        )

    @classmethod
    def none_of(cls, values: list[ProcessingStatusType]) -> ProcessingStatus:
        return ProcessingStatus(
            subject="processing_status",
            matcher=NoneOf[ProcessingStatusType](values=values),
        )


class ClassProperty(SubjectFilter):
    subject: Literal["class_property"] = "class_property"
    matcher: Union[AnyOf[str], AllOf[str], NoneOf[str]]

    @classmethod
    def any_of(cls, values: list[str]) -> ClassProperty:
        return ClassProperty(
            subject="class_property", matcher=AnyOf[str](values=values)
        )

    @classmethod
    def all_of(cls, values: list[str]) -> ClassProperty:
        return ClassProperty(
            subject="class_property", matcher=AllOf[str](values=values)
        )

    @classmethod
    def none_of(cls, values: list[str]) -> ClassProperty:
        return ClassProperty(
            subject="class_property", matcher=NoneOf[str](values=values)
        )


class ClassPropertyValue(SubjectFilter):
    subject: Literal["class_property_value"] = "class_property_value"
    matcher: Union[AnyOf[str], AllOf[str], NoneOf[str]]

    @classmethod
    def any_of(cls, values: list[str]) -> ClassPropertyValue:
        return ClassPropertyValue(
            subject="class_property_value", matcher=AnyOf[str](values=values)
        )

    @classmethod
    def all_of(cls, values: list[str]) -> ClassPropertyValue:
        return ClassPropertyValue(
            subject="class_property_value", matcher=AllOf[str](values=values)
        )

    @classmethod
    def none_of(cls, values: list[str]) -> ClassPropertyValue:
        return ClassPropertyValue(
            subject="class_property_value", matcher=NoneOf[str](values=values)
        )


class UpdatedAt(SubjectFilter):
    subject: Literal["updated_at"] = "updated_at"
    matcher: DateRange

    @classmethod
    def between(cls, start: datetime, end: datetime) -> UpdatedAt:
        return UpdatedAt(subject="updated_at", matcher=DateRange(start=start, end=end))

    @classmethod
    def before(cls, end: datetime) -> UpdatedAt:
        return UpdatedAt(subject="updated_at", matcher=DateRange(end=end))

    @classmethod
    def after(cls, start: datetime) -> UpdatedAt:
        return UpdatedAt(subject="updated_at", matcher=DateRange(start=start))


class WorkflowStatus(SubjectFilter):
    subject: Literal["workflow_status"] = "workflow_status"
    matcher: Union[AnyOf[WorkflowStatusType], NoneOf[WorkflowStatusType]]

    @classmethod
    def any_of(cls, values: list[WorkflowStatusType]) -> WorkflowStatus:
        return WorkflowStatus(
            subject="workflow_status", matcher=AnyOf[WorkflowStatusType](values=values)
        )

    @classmethod
    def none_of(cls, values: list[WorkflowStatusType]) -> WorkflowStatus:
        return WorkflowStatus(
            subject="workflow_status", matcher=NoneOf[WorkflowStatusType](values=values)
        )


class WorkflowStage(SubjectFilter):
    subject: Literal["workflow_stage"] = "workflow_stage"
    matcher: Union[AnyOf[str], NoneOf[str]]

    @classmethod
    def any_of(cls, values: list[str]) -> WorkflowStage:
        return WorkflowStage(
            subject="workflow_stage", matcher=AnyOf[str](values=values)
        )

    @classmethod
    def none_of(cls, values: list[str]) -> WorkflowStage:
        return WorkflowStage(
            subject="workflow_stage", matcher=NoneOf[str](values=values)
        )


class GroupFilter(BaseModel):
    conjunction: Literal["and", "or"] = "and"
    filters: List[Union[GroupFilter, SubjectFilter]]

    @field_validator("filters")
    def validate_filters(
        cls, value: List[GroupFilter | SubjectFilter]
    ) -> List[GroupFilter | SubjectFilter]:
        if len(value) < 2:
            raise ValueError("Must provide at least two filters.")
        return value

    def __and__(self, other: GroupFilter | SubjectFilter) -> GroupFilter:
        if isinstance(other, GroupFilter):
            if self.conjunction == "and" and other.conjunction == "and":
                return GroupFilter(
                    conjunction="and", filters=[*self.filters, *other.filters]
                )
            return GroupFilter(conjunction="and", filters=[self, other])
        if isinstance(other, SubjectFilter):
            if self.conjunction == "and":
                return GroupFilter(conjunction="and", filters=[*self.filters, other])
            return GroupFilter(conjunction="and", filters=[self, other])

    def __or__(self, other: GroupFilter | SubjectFilter) -> GroupFilter:
        if isinstance(other, GroupFilter):
            if self.conjunction == "or" and other.conjunction == "or":
                return GroupFilter(
                    conjunction="or", filters=[*self.filters, *other.filters]
                )
            return GroupFilter(conjunction="or", filters=[self, other])
        if isinstance(other, SubjectFilter):
            if self.conjunction == "or":
                return GroupFilter(conjunction="or", filters=[*self.filters, other])
            return GroupFilter(conjunction="or", filters=[self, other])



from __future__ import annotations

from typing import Dict, List

from darwin.future.core.datasets import list_datasets
from darwin.future.core.types.query import Query, QueryFilter
from darwin.future.meta.objects.dataset import Dataset


class DatasetQuery(Query[Dataset]):
    """
    DatasetQuery object with methods to manage filters, retrieve data, and execute
    filters

    Methods
    -------

    collect: Executes the query and returns the filtered data
    """

    def _collect(self) -> Dict[int, Dataset]:
        datasets, exceptions = list_datasets(self.client)
        if exceptions:
            # TODO: print and or raise exceptions, tbd how we want to handle this
            pass
        datasets_meta = [
            Dataset(client=self.client, element=dataset, meta_params=self.meta_params)
            for dataset in datasets
        ]
        if not self.filters:
            self.filters = []

        for filter in self.filters:
            datasets_meta = self._execute_filters(datasets_meta, filter)

        return dict(enumerate(datasets_meta))

    def _execute_filters(
        self, datasets: List[Dataset], filter: QueryFilter
    ) -> List[Dataset]:
        """
        Executes filtering on the local list of datasets, applying special logic for
        role filtering otherwise calls the parent method for general filtering on the
        values of the datasets

        Parameters
        ----------
        datasets : List[Dataset]
        filter : QueryFilter

        Returns
        -------
        List[Dataset]: Filtered subset of datasets
        """

        if filter.name == "releases":
            return [
                d
                for d in datasets
                if d._element is not None
                and d._element.releases
                and filter.param in [str(r) for r in d._element.releases]
            ]

        return super()._generic_execute_filter(datasets, filter)



import responses

from darwin.future.core.client import ClientCore
from darwin.future.meta.objects.dataset import Dataset
from darwin.future.meta.queries.dataset import DatasetQuery
from darwin.future.tests.core.fixtures import *


def test_dataset_collects_basic(
    base_client: ClientCore, base_datasets_json: dict
) -> None:
    query = DatasetQuery(base_client)
    with responses.RequestsMock() as rsps:
        endpoint = base_client.config.api_endpoint + "datasets"
        rsps.add(responses.GET, endpoint, json=base_datasets_json)
        datasets = query._collect().values()
        assert len(datasets) == 2
        assert all(isinstance(dataset, Dataset) for dataset in datasets)


def test_datasetquery_only_passes_back_correctly_formed_objects(
    base_client: ClientCore, base_dataset_json: dict
) -> None:
    query = DatasetQuery(base_client)
    with responses.RequestsMock() as rsps:
        endpoint = base_client.config.api_endpoint + "datasets"
        rsps.add(responses.GET, endpoint, json=[base_dataset_json, {}])
        datasets = query._collect()

        assert len(datasets) == 1
        assert isinstance(datasets[0], Dataset)


def test_dataset_filters_name(
    base_client: ClientCore, base_datasets_json: dict
) -> None:
    with responses.RequestsMock() as rsps:
        query = DatasetQuery(base_client).where(
            {"name": "name", "param": "test dataset 1"}
        )
        endpoint = base_client.config.api_endpoint + "datasets"
        rsps.add(responses.GET, endpoint, json=base_datasets_json)
        datasets = query._collect()

        assert len(datasets) == 1
        assert datasets[0]._element.slug == "test-dataset-1"


def test_dataset_filters_id(base_client: ClientCore, base_datasets_json: dict) -> None:
    with responses.RequestsMock() as rsps:
        query = DatasetQuery(base_client).where({"name": "id", "param": 1})
        endpoint = base_client.config.api_endpoint + "datasets"
        rsps.add(responses.GET, endpoint, json=base_datasets_json)
        datasets = query._collect()

        assert len(datasets) == 1
        assert datasets[0]._element.slug == "test-dataset-1"


def test_dataset_filters_slug(
    base_client: ClientCore, base_datasets_json: dict
) -> None:
    with responses.RequestsMock() as rsps:
        query = DatasetQuery(base_client).where(
            {"name": "slug", "param": "test-dataset-1"}
        )
        endpoint = base_client.config.api_endpoint + "datasets"
        rsps.add(responses.GET, endpoint, json=base_datasets_json)
        datasets = query._collect()

        assert len(datasets) == 1
        assert datasets[0]._element.slug == "test-dataset-1"


def test_dataset_filters_releases(
    base_client: ClientCore, base_datasets_json_with_releases: dict
) -> None:
    with responses.RequestsMock() as rsps:
        query = DatasetQuery(base_client).where(
            {"name": "releases", "param": "release1"}
        )
        endpoint = base_client.config.api_endpoint + "datasets"
        rsps.add(responses.GET, endpoint, json=base_datasets_json_with_releases)

        datasets_odd_ids = query._collect()

        assert len(datasets_odd_ids) == 2
        assert datasets_odd_ids[0]._element.slug == "test-dataset-1"
        assert datasets_odd_ids[1]._element.slug == "test-dataset-3"

        query2 = DatasetQuery(base_client).where(
            {"name": "releases", "param": "release2"}
        )
        datasets_even_ids = query2._collect()

        assert len(datasets_even_ids) == 2
        assert datasets_even_ids[0]._element.slug == "test-dataset-2"
        assert datasets_even_ids[1]._element.slug == "test-dataset-4"



from typing import Dict

import pytest
import responses

from darwin.future.core.client import ClientCore
from darwin.future.core.items.set_item_layout import set_item_layout
from darwin.future.data_objects.item import ItemLayout
from darwin.future.data_objects.typing import UnknownType
from darwin.future.exceptions import BadRequest
from darwin.future.tests.core.fixtures import *
from darwin.future.tests.core.items.fixtures import *


@responses.activate
def test_set_item_layout_returns_blank_object(
    base_client: ClientCore, base_layout: ItemLayout
) -> None:
    dataset_ids = [1, 2, 3]
    params = {"param1": "value1", "param2": "value2"}
    team_slug = "my_team"
    responses.add(
        responses.POST,
        f"{base_client.config.api_endpoint}v2/teams/{team_slug}/items/layout",
        json={"affected_item_count": 2},
        status=200,
    )

    response = set_item_layout(base_client, team_slug, dataset_ids, base_layout, params)

    assert response == {"affected_item_count": 2}


@responses.activate
def test_set_item_layout_raises_on_incorrect_parameters(
    base_client: ClientCore, base_layout: ItemLayout
) -> None:
    team_slug = "my_team"
    dataset_ids = [1, 2, 3]
    params: Dict[str, UnknownType] = {}

    with pytest.raises(AssertionError):
        set_item_layout(base_client, team_slug, dataset_ids, base_layout, params)


@responses.activate
def test_set_item_layout_raises_on_4xx_status_code(
    base_client: ClientCore, base_layout: ItemLayout
) -> None:
    team_slug = "my_team"
    dataset_ids = [1, 2, 3]
    params = {"param1": "value1", "param2": "value2"}

    responses.add(
        responses.POST,
        f"{base_client.config.api_endpoint}v2/teams/{team_slug}/items/layout",
        json={"error": "Bad Request"},
        status=400,
    )

    with pytest.raises(BadRequest):
        set_item_layout(base_client, team_slug, dataset_ids, base_layout, params)


@responses.activate
def test_set_item_layout_sends_correct_payload(
    base_client: ClientCore, base_layout: ItemLayout
) -> None:
    team_slug = "my_team"
    dataset_ids = [1, 2, 3]
    params = {"param1": "value1", "param2": "value2"}

    responses.add(
        responses.POST,
        f"{base_client.config.api_endpoint}v2/teams/{team_slug}/items/layout",
        json={},
        status=200,
        match=[
            responses.json_params_matcher(
                {
                    "filters": {"dataset_ids": dataset_ids, **params},
                    "layout": dict(base_layout),
                }
            )
        ],
    )

    set_item_layout(base_client, team_slug, dataset_ids, base_layout, params)



from typing import List, Literal, Optional, Tuple

import pytest

from darwin.future.data_objects.item import ItemSlot, validate_no_slashes
from darwin.future.data_objects.typing import UnknownType


def generate_extension_expectations(
    extension: str, expectation: Optional[Literal["image", "video", "pdf", "dicom"]]
) -> List[Tuple[str, Optional[Literal["image", "video", "pdf", "dicom"]]]]:
    """
    Generate a list of tuples of the form (file_name, expectation) where
    """
    return [
        (f"file.{extension}", expectation),
        (f"file.with.dots.{extension}", expectation),
        (f"/file/with/slashes.{extension}", expectation),
        (f"file/with/slashes.{extension}", expectation),
    ]


expectations_list = [
    # Supported images
    *generate_extension_expectations("jpg", "image"),
    *generate_extension_expectations("jpeg", "image"),
    *generate_extension_expectations("png", "image"),
    *generate_extension_expectations("gif", "image"),
    *generate_extension_expectations("bmp", "image"),
    # Supported documents
    *generate_extension_expectations("pdf", "pdf"),
    # Supported medical imaging
    *generate_extension_expectations("dcm", "dicom"),
    *generate_extension_expectations("nii", "dicom"),
    *generate_extension_expectations("nii.gz", "dicom"),
    # Supported videos
    *generate_extension_expectations("mp4", "video"),
    *generate_extension_expectations("avi", "video"),
    *generate_extension_expectations("mov", "video"),
    *generate_extension_expectations("wmv", "video"),
    *generate_extension_expectations("mkv", "video"),
    # Unsupported
    *generate_extension_expectations("unsupported", None),
]


class TestValidateNoSlashes:
    @pytest.mark.parametrize(
        "string", [("validname"), ("valid-name"), ("valid_name_still")]
    )
    def test_happy_paths(self, string: str) -> None:
        assert validate_no_slashes(string) == string

    @pytest.mark.parametrize("string", [(""), (123), ("/invalid_string")])
    def test_sad_paths(self, string: UnknownType) -> None:
        with pytest.raises(AssertionError):
            validate_no_slashes(string)


class TestSlotNameValidator:
    @pytest.mark.parametrize(
        "string", [("validname"), ("valid/name"), ("valid/name/still")]
    )
    def test_happy_paths(self, string: str) -> None:
        assert ItemSlot.validate_slot_name(string) == string

    @pytest.mark.parametrize("string", [(""), (123)])
    def test_sad_paths(self, string: UnknownType) -> None:
        with pytest.raises(AssertionError):
            ItemSlot.validate_slot_name(string)


class TestFpsValidator:
    def test_sets_value_if_absent(self) -> None:
        assert ItemSlot.validate_fps({}) == {"fps": 0}

    @pytest.mark.parametrize("fps", [(0), (1), (1.0), ("native")])
    def test_happy_paths(self, fps: UnknownType) -> None:
        assert ItemSlot.validate_fps({"fps": fps}) == {"fps": fps}

    @pytest.mark.parametrize("fps", [(-1), ("invalid")])
    def test_sad_paths(self, fps: UnknownType) -> None:
        with pytest.raises(AssertionError):
            ItemSlot.validate_fps({"fps": fps})


class TestRootValidator:
    @pytest.mark.parametrize("file_name, expectation", expectations_list)
    def test_happy_paths(self, file_name: str, expectation: str) -> None:
        assert (
            ItemSlot.infer_type({"file_name": file_name}).get("type") == expectation
        ), f"Failed for {file_name}, got {expectation}"



from typing import List, Tuple

from pydantic import ValidationError

from darwin.future.core.client import ClientCore
from darwin.future.data_objects.dataset import DatasetCore


def list_datasets(
    api_client: ClientCore,
) -> Tuple[List[DatasetCore], List[ValidationError]]:
    """
    Returns a list of datasets for the given team

    Parameters
    ----------
    api_client : Client
        The client to use to make the request
    team_slug : Optional[TeamSlug]
        The slug of the team to retrieve datasets for

    Returns
    -------
    List[DatasetList]:
        A list of datasets
    List[ValidationError]
        A list of Validation errors on failed objects
    """
    datasets: List[DatasetCore] = []
    errors: List[ValidationError] = []

    response = api_client.get("/datasets")
    try:
        for item in response:
            assert isinstance(item, dict)
            datasets.append(DatasetCore.model_validate(item))
    except ValidationError as e:
        errors.append(e)

    return datasets, errors



import responses

from darwin.future.core.client import ClientCore
from darwin.future.core.properties import update_property, update_property_value
from darwin.future.data_objects.properties import FullProperty, PropertyValue
from darwin.future.tests.core.fixtures import *


@responses.activate
def test_update_property(
    base_client: ClientCore, base_property_object: FullProperty
) -> None:
    # Mocking the response using responses library
    responses.add(
        responses.PUT,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/properties/{base_property_object.id}",
        json=base_property_object.model_dump(mode="json"),
        status=200,
    )
    # Call the function being tested
    property = update_property(
        base_client,
        params=base_property_object,
        team_slug=base_client.config.default_team,
    )

    # Assertions
    assert isinstance(property, FullProperty)
    assert property == base_property_object


@responses.activate
def test_update_property_value(
    base_client: ClientCore, base_property_object: FullProperty
) -> None:
    # Mocking the response using responses library
    item_id = base_property_object.id
    assert item_id
    assert base_property_object.property_values
    pv = base_property_object.property_values[0]
    responses.add(
        responses.PUT,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/properties/{item_id}/property_values/{pv.id}",
        json=pv.model_dump(),
        status=200,
    )
    # Call the function being tested
    property_value = update_property_value(
        base_client,
        params=pv,
        item_id=item_id,
        team_slug=base_client.config.default_team,
    )

    # Assertions
    assert isinstance(property_value, PropertyValue)
    assert property_value == pv



from typing import Dict, Optional


def get_item_count(dataset_dict: Dict) -> int:
    """
    Returns the number of items in the dataset.

    Parameters
    ----------
    dataset_dict: Dict
        The dataset dictionary.

    Returns
    -------
    int
        The number of items in the dataset.
    """
    num_items: Optional[int] = dataset_dict.get("num_items")
    num_videos: Optional[int] = dataset_dict.get("num_videos")
    num_images: Optional[int] = dataset_dict.get("num_images")

    if num_items is not None:
        return num_items

    return (num_images or 0) + (num_videos or 0)



from uuid import UUID

from darwin.future.meta.objects.base import MetaBase


class V7ID(MetaBase[UUID]):
    @property
    def id(self) -> UUID:
        return self._element

    def __str__(self) -> str:
        return str(self._element)

    def __repr__(self) -> str:
        return str(self)



from pydantic import BaseModel, ConfigDict


class DefaultDarwin(BaseModel):
    """
    Default Darwin-Py pydantic settings for meta information.
    Default settings include:
        - auto validating variables on setting/assignment
        - underscore attributes are private
        - objects are passed by reference to prevent unnecesary data copying
    """

    # TODO[pydantic]: The following keys were removed: `underscore_attrs_are_private`, `copy_on_model_validation`.
    # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-config for more information.
    model_config = ConfigDict(validate_assignment=True)



import responses
from pydantic import ValidationError
from pytest import raises

from darwin.future.core.client import ClientCore
from darwin.future.core.types.common import JSONType
from darwin.future.core.workflows import get_workflow
from darwin.future.data_objects.workflow import WorkflowCore
from darwin.future.exceptions import BadRequest
from darwin.future.tests.core.fixtures import *


@responses.activate
def test_get_workflow(
    base_client: ClientCore, base_single_workflow_object: JSONType
) -> None:
    # Mocking the response using responses library
    response_data = base_single_workflow_object
    workflow_id = "1"
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/workflows/{workflow_id}",
        json=response_data,
        status=200,
    )

    # Call the function being tested
    workflow = get_workflow(base_client, workflow_id)

    # Assertions
    assert isinstance(workflow, WorkflowCore)


@responses.activate
def test_get_workflow_with_team_slug(
    base_client: ClientCore, base_single_workflow_object: JSONType
) -> None:
    # Mocking the response using responses library
    team_slug = "team-slug"
    workflow_id = "1"

    response_data = base_single_workflow_object
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{team_slug}/workflows/{workflow_id}",
        json=response_data,
        status=200,
    )

    # Call the function being tested
    workflow = get_workflow(base_client, workflow_id, team_slug)

    # Assertions
    assert isinstance(workflow, WorkflowCore)


@responses.activate
def test_get_workflows_with_invalid_response(base_client: ClientCore) -> None:
    # Mocking the response using responses library
    # fmt: off
    NON_EXISTENT_ID = "1"
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/workflows/{NON_EXISTENT_ID}",
        json={'invalid': 'response'},
        status=200
    )
    # fmt: on

    # Call the function being tested
    with raises(ValidationError):
        get_workflow(base_client, NON_EXISTENT_ID)


@responses.activate
def test_get_workflows_with_error(base_client: ClientCore) -> None:
    # Mocking the response using responses library
    # fmt: off
    NON_EXISTENT_ID = "1"
    responses.add(
        responses.GET,
        f"{base_client.config.base_url}api/v2/teams/{base_client.config.default_team}/workflows/{NON_EXISTENT_ID}",
        json="{}",
        status=400
    )
    # fmt: on
    with raises(BadRequest):
        get_workflow(base_client, NON_EXISTENT_ID)



from typing import List
from uuid import UUID, uuid4

import pytest
import responses
from responses.matchers import query_param_matcher

from darwin.future.core.client import ClientCore
from darwin.future.data_objects.page import Page
from darwin.future.meta.queries.item_id import ItemIDQuery
from darwin.future.tests.core.fixtures import *


@pytest.fixture
def base_ItemIDQuery(base_client: ClientCore) -> ItemIDQuery:
    return ItemIDQuery(
        base_client, meta_params={"dataset_id": 0000, "team_slug": "test_team"}
    )


@pytest.fixture
def list_of_uuids() -> List[UUID]:
    return [uuid4() for _ in range(10)]


def test_pagination_collects_all(
    base_client: ClientCore, base_ItemIDQuery: ItemIDQuery, list_of_uuids: List[UUID]
) -> None:
    base_ItemIDQuery.page = Page(size=5)
    team_slug = base_ItemIDQuery.meta_params["team_slug"]
    dataset_id = base_ItemIDQuery.meta_params["dataset_id"]
    str_ids = [str(uuid) for uuid in list_of_uuids]
    with responses.RequestsMock() as rsps:
        endpoint = (
            base_client.config.api_endpoint + f"v2/teams/{team_slug}/items/list_ids"
        )
        rsps.add(
            responses.GET,
            endpoint,
            match=[
                query_param_matcher(
                    {
                        "page[offset]": "0",
                        "page[size]": "5",
                        "dataset_ids": str(dataset_id),
                    }
                )
            ],
            json={"item_ids": [str(uuid) for uuid in str_ids[:5]]},
        )
        rsps.add(
            responses.GET,
            endpoint,
            match=[
                query_param_matcher(
                    {
                        "page[offset]": "5",
                        "page[size]": "5",
                        "dataset_ids": str(dataset_id),
                    }
                )
            ],
            json={"item_ids": [str(uuid) for uuid in str_ids[5:]]},
        )
        rsps.add(
            responses.GET,
            endpoint,
            match=[
                query_param_matcher(
                    {
                        "page[offset]": "10",
                        "page[size]": "5",
                        "dataset_ids": str(dataset_id),
                    }
                )
            ],
            json={"item_ids": []},
        )

        ids = base_ItemIDQuery.collect_all()
        raw_ids = [x.id for x in ids]
        assert len(rsps.calls) == 3
        assert len(ids) == 10
        assert raw_ids == list_of_uuids
        assert base_ItemIDQuery.page.offset == 10
        assert base_ItemIDQuery.completed is True


def test_iterable_collects_all(
    base_client: ClientCore, base_ItemIDQuery: ItemIDQuery, list_of_uuids: List[UUID]
) -> None:
    base_ItemIDQuery.page = Page(size=5)
    team_slug = base_ItemIDQuery.meta_params["team_slug"]
    dataset_id = base_ItemIDQuery.meta_params["dataset_id"]
    str_ids = [str(uuid) for uuid in list_of_uuids]
    with responses.RequestsMock() as rsps:
        endpoint = (
            base_client.config.api_endpoint + f"v2/teams/{team_slug}/items/list_ids"
        )
        rsps.add(
            responses.GET,
            endpoint,
            match=[
                query_param_matcher(
                    {
                        "page[offset]": "0",
                        "page[size]": "5",
                        "dataset_ids": str(dataset_id),
                    }
                )
            ],
            json={"item_ids": [str(uuid) for uuid in str_ids[:5]]},
        )
        rsps.add(
            responses.GET,
            endpoint,
            match=[
                query_param_matcher(
                    {
                        "page[offset]": "5",
                        "page[size]": "5",
                        "dataset_ids": str(dataset_id),
                    }
                )
            ],
            json={"item_ids": [str(uuid) for uuid in str_ids[5:]]},
        )
        rsps.add(
            responses.GET,
            endpoint,
            match=[
                query_param_matcher(
                    {
                        "page[offset]": "10",
                        "page[size]": "5",
                        "dataset_ids": str(dataset_id),
                    }
                )
            ],
            json={"item_ids": []},
        )

        ids = base_ItemIDQuery
        for i, item in enumerate(ids):
            if i < 5:
                assert item.id in list_of_uuids[:5]
                assert len(rsps.calls) == 1
            elif i < 10:
                assert item.id in list_of_uuids[:10]
                assert len(rsps.calls) == 2

        assert len(rsps.calls) == 3
        assert base_ItemIDQuery.page.offset == 10
        assert base_ItemIDQuery.completed is True
        assert len(ids) == 10


def test_can_become_iterable(
    base_client: ClientCore, base_ItemIDQuery: ItemIDQuery, list_of_uuids: List[UUID]
) -> None:
    base_ItemIDQuery.page = Page(size=20)
    team_slug = base_ItemIDQuery.meta_params["team_slug"]
    dataset_id = base_ItemIDQuery.meta_params["dataset_id"]
    str_ids = [str(uuid) for uuid in list_of_uuids]
    with responses.RequestsMock() as rsps:
        endpoint = (
            base_client.config.api_endpoint + f"v2/teams/{team_slug}/items/list_ids"
        )
        rsps.add(
            responses.GET,
            endpoint,
            match=[
                query_param_matcher(
                    {
                        "page[offset]": "0",
                        "page[size]": "20",
                        "dataset_ids": str(dataset_id),
                    }
                )
            ],
            json={"item_ids": [str(uuid) for uuid in str_ids]},
        )

        ids = list(base_ItemIDQuery)
        ids_raw = [x.id for x in ids]
        assert len(rsps.calls) == 1
        assert ids_raw == list_of_uuids


def test_get_specific_index_collects_correct_page(
    base_client: ClientCore, base_ItemIDQuery: ItemIDQuery, list_of_uuids: List[UUID]
) -> None:
    base_ItemIDQuery.page = Page(size=5)
    team_slug = base_ItemIDQuery.meta_params["team_slug"]
    dataset_id = base_ItemIDQuery.meta_params["dataset_id"]
    str_ids = [str(uuid) for uuid in list_of_uuids]
    with responses.RequestsMock() as rsps:
        endpoint = (
            base_client.config.api_endpoint + f"v2/teams/{team_slug}/items/list_ids"
        )
        rsps.add(
            responses.GET,
            endpoint,
            match=[
                query_param_matcher(
                    {
                        "page[offset]": "5",
                        "page[size]": "5",
                        "dataset_ids": str(dataset_id),
                    }
                )
            ],
            json={"item_ids": [str(uuid) for uuid in str_ids[5:]]},
        )
        base_ItemIDQuery[7]


def test_sort_method(base_client: ClientCore):
    item_id_query = ItemIDQuery(
        base_client, meta_params={"dataset_id": 0000, "team_slug": "test_team"}
    )

    item_id_query.sort(accuracy="desc", byte_size="asc")

    assert len(item_id_query.filters) == 2
    assert item_id_query.filters[0].name == "sort[accuracy]"
    assert item_id_query.filters[0].param == "desc"



import gzip
import json
import os
import re
import subprocess
from pathlib import Path
from typing import Dict, List, Optional, Tuple

from rich.console import Console

console = Console()


def _check_ffmpeg_version():
    """
    Check if FFmpeg version 5 or higher is installed.
    Raises RuntimeError if FFmpeg is not found or version is lower.
    """
    try:
        result = subprocess.run(["ffmpeg", "-version"], capture_output=True, text=True)
        version_line = result.stdout.split("\n")[0]
        # Extract major version number (e.g., "ffmpeg version 5.1.2" -> "5")
        version_match = re.search(r"ffmpeg version (\d+)", version_line)
        if not version_match:
            raise RuntimeError("Could not determine FFmpeg version")

        major_version = int(version_match.group(1))
        if major_version < 5:
            raise RuntimeError(
                f"FFmpeg version 5 or higher required, found version {major_version}"
            )

    except FileNotFoundError:
        raise RuntimeError(
            "FFmpeg not found. Please install FFmpeg version 5 or higher"
        )


def _create_directories(base_dir: str) -> Dict[str, str]:
    """Create required directory structure for artifacts"""

    paths = {
        "base_dir": base_dir,
        "segments": os.path.join(base_dir, "segments"),
        "sections": os.path.join(base_dir, "sections"),
    }

    # Create high/low quality segment dirs
    paths["segments_high"] = os.path.join(paths["segments"], "high")
    paths["segments_low"] = os.path.join(paths["segments"], "low")

    for path in paths.values():
        os.makedirs(path, exist_ok=True)

    return paths


def _maybe_repair_video(source_file: str, output_dir: str) -> Tuple[bool, str]:
    """
    Attempt to repair video if errors are detected.

    Args:
        source_file: Path to source video file
        output_dir: Directory to store repaired video if needed

    Returns:
        Tuple[bool, str]: (was_repaired, final_source_file_path)
    """
    console.print(f"Checking video for errors: {source_file}")
    errors = _check_video_for_errors(source_file)
    if errors:
        errors_list = errors.split("\n")
        first_three = "\n".join(errors_list[:3])
        console.print(f"Video contains errors:\n{first_three}\n...")
        console.print("Attempting to repair video...")
        repaired_file = _attempt_video_repair(source_file, output_dir)
        console.print(f"Video repaired successfully: {repaired_file}")
        return (True, repaired_file)
    else:
        console.print("No errors detected, proceeding with original video")
        return (False, source_file)


def _check_video_for_errors(source_file: str) -> str:
    """
    Check if video file has any errors using FFmpeg error detection.

    Args:
        source_file: Path to source video file

    Returns:
        str: Error message if errors were detected, empty string otherwise
    """
    cmd = [
        "ffmpeg",
        "-err_detect",
        "explode",
        "-xerror",
        "-v",
        "error",
        "-i",
        source_file,
        "-f",
        "null",
        "-",
    ]

    try:
        result = subprocess.run(cmd, capture_output=True, text=True, check=True)
        return result.stderr.strip()
    except subprocess.CalledProcessError as e:
        return e.stderr.strip()


def _attempt_video_repair(source_file: str, output_dir: str) -> str:
    """
    Attempt to repair corrupted video by re-encoding it using hardware acceleration if available.

    Args:
        source_file: Path to source video file
        output_dir: Directory to store repaired video

    Returns:
        str: Path to repaired video file
    """
    output_file = os.path.join(output_dir, f"repaired_{os.path.basename(source_file)}")

    try:
        # First try hardware accelerated encoding (H265)
        cmd = [
            "ffmpeg",
            "-y",
            "-fflags",
            "discardcorrupt+genpts",
            "-vaapi_device",
            "/dev/dri/renderD128",
            "-i",
            source_file,
            "-vf",
            "format=nv12,hwupload",
            "-c:v",
            "hevc_vaapi",
            "-c:a",
            "copy",
            "-vsync",
            "cfr",
            output_file,
        ]
        subprocess.run(cmd, check=True, capture_output=True)
        return output_file
    except subprocess.CalledProcessError:
        console.print(
            "Hardware acceleration with H265 failed, falling back to software encoding with H264"
        )

        cmd = [
            "ffmpeg",
            "-y",
            "-fflags",
            "discardcorrupt+genpts",
            "-i",
            source_file,
            "-c:v",
            "libx264",
            "-c:a",
            "copy",
            "-vsync",
            "cfr",
            output_file,
        ]
        subprocess.run(cmd, check=True, capture_output=True)
        return output_file


def _count_frames(source_file: str) -> int:
    cmd = [
        "ffprobe",
        "-v",
        "error",
        "-select_streams",
        "v:0",
        "-count_frames",
        "-show_entries",
        "stream=nb_read_frames",
        "-of",
        "json",
        source_file,
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    return int(json.loads(result.stdout)["streams"][0]["nb_read_frames"])


def _get_video_metadata(source_file: str) -> Dict:
    """Extract video metadata using ffprobe"""
    cmd = [
        "ffprobe",
        "-v",
        "error",
        "-select_streams",
        "v:0",
        "-show_entries",
        "stream=width,height,avg_frame_rate,duration",
        "-of",
        "json",
        source_file,
    ]

    result = subprocess.run(cmd, capture_output=True, text=True)
    data = json.loads(result.stdout)["streams"][0]

    # Try to calculate native fps from avg_frame_rate first
    try:
        num, den = map(int, data["avg_frame_rate"].split("/"))
        if den > 0:
            native_fps = num / den
        else:
            native_fps = 0.0
    except (ValueError, ZeroDivisionError):
        native_fps = 0.0

    # If avg_frame_rate calculation failed, try calculating from frame count and duration
    if native_fps == 0.0:
        duration = float(data["duration"])
        total_frames = _count_frames(source_file)
        native_fps = round(total_frames / duration, 2)

    return {
        "width": int(data["width"]),
        "height": int(data["height"]),
        "native_fps": native_fps,
    }


def _calculate_avg_bitrate(index_data: str, segments: List[str]) -> Optional[float]:
    """
    Calculate average bitrate from HLS segments.
    Returns None if no valid segments found.
    """
    bitrates = []

    # Parse durations from index file
    durations = []
    for line in index_data.splitlines():
        if line.startswith("#EXTINF:"):
            try:
                duration = float(line.split(":")[1].split(",")[0])
                durations.append(duration)
            except (IndexError, ValueError):
                continue

    # Calculate bitrates for each segment
    for duration, segment in zip(durations, segments):
        try:
            if duration > 0:
                size = os.path.getsize(segment)
                # Convert bytes to bits and divide by duration to get bits per second
                bitrate = (size * 8) / duration
                bitrates.append(bitrate)
        except (OSError, ZeroDivisionError):
            continue

    # Calculate average bitrate if we have valid measurements
    if bitrates:
        return sum(bitrates) / len(bitrates)
    return None


def _extract_segments(source_file: str, dirs: Dict, segment_length: int) -> Dict:
    """
    Extract HLS segments in high and low quality
    Returns segment info and frame counts per segment
    """
    qualities = {
        "high": {"crf": 23, "gop": 15},
        "low": {
            "crf": 40,
            "gop": 15,
            "scale": "-2:'if(gt(ih,720),max(ceil(ih/4)*2,720),ih)'",
        },
    }

    bitrates = {}

    for quality, opts in qualities.items():
        quality_dir = dirs["segments_" + quality]
        segment_pattern = os.path.join(quality_dir, "%09d.ts")
        index_path = os.path.join(quality_dir, "index.m3u8")

        # Build ffmpeg command
        cmd = [
            "ffmpeg",
            "-hide_banner",
            "-v",
            "error",
            "-i",
            source_file,
            "-c:v",
            "libx264",
            "-crf",
            str(opts["crf"]),
            "-g",
            str(opts["gop"]),
            "-f",
            "hls",
            "-hls_time",
            str(segment_length),
            "-hls_list_size",
            "0",
            "-start_number",
            "0",
            "-hls_segment_filename",
            segment_pattern,
            "-vsync",
            "passthrough",
            "-max_muxing_queue_size",
            "1024",
        ]

        # Add scale filter for low quality
        if "scale" in opts:
            cmd.extend(["-vf", f"scale={opts['scale']}"])

        cmd.append(index_path)

        subprocess.run(cmd, check=True)

        # Read index file and calculate bitrate
        with open(index_path) as f:
            index_data = f.read()
            segments = sorted(Path(quality_dir).glob("*.ts"))
            bitrate = _calculate_avg_bitrate(index_data, [str(s) for s in segments])
            bitrates[quality] = bitrate

    return {"bitrates": bitrates}


def _extract_thumbnail(source_file: str, output_path: str, total_frames: int) -> str:
    """Extract thumbnail from middle frame"""
    middle_frame = total_frames // 2

    cmd = [
        "ffmpeg",
        "-hide_banner",
        "-v",
        "error",
        "-i",
        source_file,
        "-vf",
        f"select=gte(n\\,{middle_frame}),scale=w=356:h=200:force_original_aspect_ratio=decrease",
        "-vframes",
        "1",
        "-y",
        output_path,
    ]

    subprocess.run(cmd, check=True)
    return output_path


def _get_frames_timestamps(source_file: str) -> List[float]:
    """Get frame timestamps using ffmpeg showinfo filter"""
    cmd = [
        "ffmpeg",
        "-hide_banner",
        "-v",
        "info",
        "-i",
        source_file,
        "-vsync",
        "passthrough",
        "-vf",
        "showinfo",
        "-f",
        "null",
        "-",
    ]

    result = subprocess.run(cmd, capture_output=True, text=True)

    # Parse timestamps from stderr
    frames = []
    for line in result.stderr.splitlines():
        if "pts_time:" in line:
            pts_time = line.split("pts_time:")[1].split()[0]
            frames.append(float(pts_time))

    return frames


def _extract_frames(source_file: str, output_dir: str, downsampling_step: float):
    """Extract frames using ffmpeg with optional downsampling"""
    frame_pattern = os.path.join(output_dir, "%09d.png")

    if downsampling_step > 1:
        # Use select filter to precisely control what frames are extracted
        # This matches the frame selection logic in the manifest
        select_expr = f"select='eq(trunc(trunc((n+1)/{downsampling_step})*{downsampling_step})\\,n)'"
        cmd = [
            "ffmpeg",
            "-hide_banner",
            "-v",
            "error",
            "-i",
            source_file,
            "-start_number",
            "0",
            "-vsync",
            "passthrough",
            "-vf",
            select_expr,
            "-f",
            "image2",
            frame_pattern,
        ]
    else:
        # Extract all frames when no downsampling needed
        cmd = [
            "ffmpeg",
            "-hide_banner",
            "-v",
            "error",
            "-i",
            source_file,
            "-start_number",
            "0",
            "-vsync",
            "passthrough",
            "-f",
            "image2",
            frame_pattern,
        ]

    subprocess.run(cmd, check=True)


def _get_segment_frame_counts(segments_dir: str) -> List[int]:
    """Get frame counts for each segment in order"""
    segments = sorted(Path(segments_dir).glob("*.ts"))
    segment_frame_counts = []
    for segment in segments:
        count = _count_frames(str(segment))
        segment_frame_counts.append(count)
    return segment_frame_counts


def _create_frames_manifest(
    source_file: str, segments_dir: str, downsampling_step: float, manifest_path: str
) -> Dict:
    """
    Create frames manifest mapping frames to segments
    Format: FRAME_NO_IN_SEGMENT:SEGMENT_NO:VISIBILITY_FLAG:TIMESTAMP
    """
    frames_timestamps = _get_frames_timestamps(source_file)
    segment_frame_counts = _get_segment_frame_counts(segments_dir)

    file_lines = []
    visible_frames = 0
    frame_no = 0
    segment_no = 0
    frame_no_segment = 0

    for frame_timestamp in frames_timestamps:
        # Check if frame should be visible based on downsampling
        next_visible_frame = int(visible_frames * downsampling_step)
        visibility = 1 if frame_no == next_visible_frame else 0

        if visibility == 1:
            visible_frames += 1

        # Move to next segment if current is full
        if frame_no_segment >= segment_frame_counts[segment_no]:
            segment_no += 1
            frame_no_segment = 0

        # Create manifest line
        line = f"{frame_no_segment}:{segment_no}:{visibility}:{frame_timestamp}"
        file_lines.append(line)

        frame_no += 1
        frame_no_segment += 1

    with open(manifest_path, "w") as f:
        f.write("\n".join(file_lines))

    return {"visible_frames": visible_frames, "total_frames": frame_no}


def _get_hls_index_with_storage_urls(segments_dir: str, storage_key_prefix: str) -> str:
    """
    Replaces relative paths in HLS index file by storage keys.
    """
    index_path = os.path.join(segments_dir, "index.m3u8")
    with open(index_path, "r") as f:
        content = f.read()
        return re.sub(
            r"^(.*\.ts)$",
            lambda m: f"{storage_key_prefix}/{m.group(1)}",
            content,
            flags=re.MULTILINE,
        )


def _maybe_extract_audio_peaks(source_file: str, output_dir: str) -> Optional[str]:
    """
    Extract audio peaks from video file and gzip the result.
    Returns path to the gzipped peaks file if audio stream exists, None otherwise.
    """
    # First check if audio stream exists
    cmd = [
        "ffprobe",
        "-v",
        "error",
        "-select_streams",
        "a",
        "-show_entries",
        "stream=codec_type",
        "-of",
        "json",
        source_file,
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    data = json.loads(result.stdout)

    if not data.get("streams"):
        console.print("No audio streams found")
        return None

    raw_output_path = os.path.join(output_dir, "audio_peaks.raw")
    gzipped_output_path = os.path.join(output_dir, "audio_peaks.gz")

    cmd = [
        "ffmpeg",
        "-hide_banner",
        "-v",
        "error",
        "-i",
        source_file,
        "-map",
        "0:a:0",
        "-ac",
        "1",
        "-af",
        "aresample=1000,asetnsamples=1",
        "-f",
        "u8",
        raw_output_path,
    ]

    try:
        subprocess.run(cmd, check=True)
        # Gzip the output file
        with open(raw_output_path, "rb") as f_in:
            with gzip.open(gzipped_output_path, "wb") as f_out:
                f_out.writelines(f_in)
        # Remove the raw file
        os.remove(raw_output_path)
        return gzipped_output_path
    except (subprocess.CalledProcessError, OSError):
        console.print("Failed to extract audio peaks")
        if os.path.exists(raw_output_path):
            os.remove(raw_output_path)
        if os.path.exists(gzipped_output_path):
            os.remove(gzipped_output_path)
        return None


def extract_artifacts(
    source_file: str,
    output_dir: str,
    storage_key_prefix: str,
    *,
    fps: float = 0.0,
    segment_length: int = 2,
    repair: bool = False,
) -> Dict:
    """
    Extracts video artifacts including segments, frames, thumbnail for
    read-only registration in the Darwin platform.

    Args:
        source_file: Path to source video file
        output_dir: Directory to store generated artifacts
        storage_key_prefix: Prefix for storage keys
        fps: Desired frames per second (0.0 for native fps), defaults to 0.0
        segment_length: Length of each segment in seconds, defaults to 2
        repair: If True, attempt to repair video if errors are detected, defaults to False

    Returns:
        Dict containing metadata and paths to generated artifacts

    Raises:
        FileNotFoundError: If source_file does not exist
    """
    if not os.path.exists(source_file):
        raise FileNotFoundError(f"Source video file not found: {source_file}")

    _check_ffmpeg_version()
    dirs = _create_directories(output_dir)

    repaired = False
    if repair:
        repaired, source_file = _maybe_repair_video(source_file, dirs["base_dir"])

    storage_key_prefix = storage_key_prefix.strip("/")

    console.print("\nExtracting video metadata...")

    metadata = _get_video_metadata(source_file)
    native_fps = float(metadata["native_fps"])
    downsampling_step = round(max(native_fps / fps, 1.0), 4) if fps > 0 else 1.0
    source_file_size = os.path.getsize(source_file)

    console.print(f"\nVideo resolution: {metadata['width']}x{metadata['height']}")
    console.print(f"Native FPS: {native_fps}")
    console.print(f"Downsampling step: {downsampling_step}")
    console.print(f"Source file size: {source_file_size} bytes")

    console.print("\nExtracting video segments...")

    segments_metadata = _extract_segments(
        source_file=source_file, dirs=dirs, segment_length=segment_length
    )

    console.print("\nExtracting frames...")

    _extract_frames(source_file, dirs["sections"], downsampling_step)

    console.print("\nCreating frames manifest...")

    manifest_metadata = _create_frames_manifest(
        source_file=source_file,
        segments_dir=dirs["segments_high"],
        downsampling_step=downsampling_step,
        manifest_path=os.path.join(dirs["base_dir"], "frames_manifest.txt"),
    )

    console.print("\nExtracting thumbnail...")

    _extract_thumbnail(
        source_file=source_file,
        output_path=os.path.join(dirs["base_dir"], "thumbnail.jpg"),
        total_frames=manifest_metadata["total_frames"],
    )

    console.print("\nExtracting audio peaks...")

    _maybe_extract_audio_peaks(source_file, dirs["base_dir"])

    console.print("\nProcessing segment indices...")

    hq_hls_index = _get_hls_index_with_storage_urls(
        dirs["segments_high"], f"{storage_key_prefix}/segments/high"
    )
    lq_hls_index = _get_hls_index_with_storage_urls(
        dirs["segments_low"], f"{storage_key_prefix}/segments/low"
    )

    console.print("\nSaving metadata...")

    source_file_name = os.path.basename(source_file)

    # Prepare final metadata
    result_metadata = {
        "repaired": repaired,
        "source_file": source_file,
        "storage_key_prefix": storage_key_prefix,
        "registration_payload": {
            "type": "video",
            "width": metadata["width"],
            "height": metadata["height"],
            "native_fps": metadata["native_fps"],
            "fps": fps,
            "visible_frames": manifest_metadata["visible_frames"],
            "total_frames": manifest_metadata["total_frames"],
            # After item registration with this payload,
            # DARWIN will use storage keys from HLS indexes and fields defined below
            # to create signed URLs and fetch files from storage.
            # Therefore, the storage key must be correct.
            "hls_segments": {
                "high_quality": {
                    "index": hq_hls_index,
                    "bitrate": segments_metadata["bitrates"]["high"],
                },
                "low_quality": {
                    "index": lq_hls_index,
                    "bitrate": segments_metadata["bitrates"]["low"],
                },
            },
            "storage_key": f"{storage_key_prefix}/{source_file_name}",
            "storage_sections_key_prefix": f"{storage_key_prefix}/sections",
            "storage_frames_manifest_key": f"{storage_key_prefix}/frames_manifest.txt",
            "storage_thumbnail_key": f"{storage_key_prefix}/thumbnail.jpg",
            "total_size_bytes": source_file_size,
            "name": source_file_name,
            "path": "/",
        },
    }

    # Add audio peaks key if audio was extracted
    # Must be uploaded with Content-Encoding gzip
    audio_peaks_path = os.path.join(dirs["base_dir"], "audio_peaks.gz")
    if os.path.exists(audio_peaks_path):
        result_metadata["registration_payload"][
            "storage_audio_peaks_key"
        ] = f"{storage_key_prefix}/audio_peaks.gz"

    with open(os.path.join(dirs["base_dir"], "metadata.json"), "w") as f:
        json.dump(result_metadata, f, indent=2)

    return result_metadata



from uuid import UUID

import pytest
import responses
from responses import json_params_matcher

from darwin.future.data_objects.item import ItemLayout, ItemSlot
from darwin.future.exceptions import BadRequest
from darwin.future.meta.objects.item import Item
from darwin.future.tests.meta.objects.fixtures import *


def test_item_properties(item: Item) -> None:
    assert isinstance(item.name, str)
    assert isinstance(item.id, UUID)
    assert isinstance(item.slots, list)
    for slot in item.slots:
        assert isinstance(slot, ItemSlot)
    assert isinstance(item.path, str)
    assert isinstance(item.dataset_id, int)
    assert isinstance(item.processing_status, str)
    assert isinstance(item.archived, (bool, type(None)))
    assert isinstance(item.priority, (int, type(None)))
    assert isinstance(item.tags, (list, dict, type(None)))
    assert isinstance(item.layout, (ItemLayout, type(None)))


def test_delete(item: Item) -> None:
    with responses.RequestsMock() as rsps:
        team_slug = item.meta_params["team_slug"]
        dataset_id = item.meta_params["dataset_id"]
        rsps.add(
            rsps.DELETE,
            item.client.config.api_endpoint + f"v2/teams/{team_slug}/items",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id)],
                            "dataset_ids": [dataset_id],
                        }
                    }
                )
            ],
            json={},
        )
        item.delete()


def test_delete_with_bad_team_slug(item: Item) -> None:
    with pytest.raises(AssertionError):
        item.meta_params["team_slug"] = 123
        item.delete()


def test_move_to_folder(item: Item) -> None:
    with responses.RequestsMock() as rsps:
        team_slug = item.meta_params["team_slug"]
        dataset_id = item.meta_params["dataset_id"]
        path = "/new_folder"
        rsps.add(
            rsps.POST,
            item.client.config.api_endpoint + f"v2/teams/{team_slug}/items/path",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id)],
                            "dataset_ids": [dataset_id],
                        },
                        "path": path,
                    }
                )
            ],
            json={},
        )
        item.move_to_folder(path)


def test_move_to_folder_raises_on_incorrect_parameters(item: Item) -> None:
    with responses.RequestsMock() as rsps:
        team_slug = item.meta_params["team_slug"]
        dataset_id = item.meta_params["dataset_id"]
        path = 1234
        rsps.add(
            rsps.POST,
            item.client.config.api_endpoint + f"v2/teams/{team_slug}/items/path",
            status=400,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id)],
                            "dataset_ids": [dataset_id],
                        },
                        "path": path,
                    }
                )
            ],
            json={},
        )
        with pytest.raises(BadRequest):
            item.move_to_folder(path)  # type: ignore


def test_move_to_folder_with_bad_team_slug(item: Item) -> None:
    with pytest.raises(AssertionError):
        path = "/new_folder"
        item.meta_params["team_slug"] = 123
        item.move_to_folder(path)


def test_set_priority(item: Item) -> None:
    with responses.RequestsMock() as rsps:
        team_slug = item.meta_params["team_slug"]
        dataset_id = item.meta_params["dataset_id"]
        priority = 10
        rsps.add(
            rsps.POST,
            item.client.config.api_endpoint + f"v2/teams/{team_slug}/items/priority",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id)],
                            "dataset_ids": [dataset_id],
                        },
                        "priority": priority,
                    }
                )
            ],
            json={},
        )
        item.set_priority(priority)


def test_set_priority_raises_on_incorrect_parameters(item: Item) -> None:
    with responses.RequestsMock() as rsps:
        team_slug = item.meta_params["team_slug"]
        dataset_id = item.meta_params["dataset_id"]
        priority = "10"
        rsps.add(
            rsps.POST,
            item.client.config.api_endpoint + f"v2/teams/{team_slug}/items/priority",
            status=400,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id)],
                            "dataset_ids": [dataset_id],
                        },
                        "priority": priority,
                    }
                )
            ],
            json={},
        )
        with pytest.raises(BadRequest):
            item.set_priority(priority)  # type: ignore


def test_set_priority_with_bad_team_slug(item: Item) -> None:
    with pytest.raises(AssertionError):
        priority = 10
        item.meta_params["team_slug"] = 123
        item.set_priority(priority)


def test_restore(item: Item) -> None:
    with responses.RequestsMock() as rsps:
        team_slug = item.meta_params["team_slug"]
        dataset_id = item.meta_params["dataset_id"]
        rsps.add(
            rsps.POST,
            item.client.config.api_endpoint + f"v2/teams/{team_slug}/items/restore",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id)],
                            "dataset_ids": [dataset_id],
                        }
                    }
                )
            ],
            json={},
        )
        item.restore()


def test_restore_with_bad_team_slug(item: Item) -> None:
    with pytest.raises(AssertionError):
        item.meta_params["team_slug"] = 123
        item.restore()


def test_archive(item: Item) -> None:
    with responses.RequestsMock() as rsps:
        team_slug = item.meta_params["team_slug"]
        dataset_id = item.meta_params["dataset_id"]
        rsps.add(
            rsps.POST,
            item.client.config.api_endpoint + f"v2/teams/{team_slug}/items/archive",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id)],
                            "dataset_ids": [dataset_id],
                        }
                    }
                )
            ],
            json={},
        )
        item.archive()


def test_archive_with_bad_team_slug(item: Item) -> None:
    with pytest.raises(AssertionError):
        item.meta_params["team_slug"] = 123
        item.archive()


def test_set_layout(item: Item) -> None:
    with responses.RequestsMock() as rsps:
        team_slug = item.meta_params["team_slug"]
        dataset_id = item.meta_params["dataset_id"]
        layout = ItemLayout(version=1, type="grid", slots=["slot1", "slot2"])
        rsps.add(
            rsps.POST,
            item.client.config.api_endpoint + f"v2/teams/{team_slug}/items/layout",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id)],
                            "dataset_ids": [dataset_id],
                        },
                        "layout": layout.model_dump(),
                    }
                )
            ],
            json={},
        )
        item.set_layout(layout)


def test_set_layout_raises_on_incorrect_parameters(item: Item) -> None:
    with responses.RequestsMock():
        item.meta_params["team_slug"]
        item.meta_params["dataset_id"]
        layout = "invalid_layout"
        with pytest.raises(AssertionError):
            item.set_layout(layout)  # type: ignore


def test_set_layout_with_bad_team_slug(item: Item) -> None:
    with pytest.raises(AssertionError):
        layout = ItemLayout(version=1, type="grid", slots=["slot1", "slot2"])
        item.meta_params["team_slug"] = 123
        item.set_layout(layout)


def test_tag(item: Item) -> None:
    with responses.RequestsMock() as rsps:
        team_slug = item.meta_params["team_slug"]
        dataset_id = item.meta_params["dataset_id"]
        tag_id = 123456
        rsps.add(
            rsps.POST,
            item.client.config.api_endpoint + f"v2/teams/{team_slug}/items/slots/tags",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id)],
                            "dataset_ids": [dataset_id],
                        },
                        "annotation_class_id": tag_id,
                    }
                )
            ],
            json={},
        )
        item.tag(tag_id)


def test_tag_bad_input(item: Item) -> None:
    with responses.RequestsMock():
        tag_id = "123456"
        with pytest.raises(BadRequest) as excinfo:
            item.tag(tag_id)  # type: ignore
        (msg,) = excinfo.value.args
        assert msg == "tag_id must be an integer, got <class 'str'>"


def test_untag(item: Item) -> None:
    with responses.RequestsMock() as rsps:
        team_slug = item.meta_params["team_slug"]
        dataset_id = item.meta_params["dataset_id"]
        tag_id = 123456
        rsps.add(
            rsps.DELETE,
            item.client.config.api_endpoint + f"v2/teams/{team_slug}/items/slots/tags",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id)],
                            "dataset_ids": [dataset_id],
                        },
                        "annotation_class_id": tag_id,
                    }
                )
            ],
            json={},
        )
        item.untag(tag_id)


def test_untag_bad_input(item: Item) -> None:
    with responses.RequestsMock():
        tag_id = "123456"
        with pytest.raises(BadRequest) as excinfo:
            item.untag(tag_id)  # type: ignore
        (msg,) = excinfo.value.args
        assert msg == "tag_id must be an integer, got <class 'str'>"


def test_set_stage(item: Item) -> None:
    with responses.RequestsMock() as rsps:
        team_slug = item.meta_params["team_slug"]
        dataset_id = item.meta_params["dataset_id"]
        stage_id = "123456"
        workflow_id = "123456"
        rsps.add(
            rsps.POST,
            item.client.config.api_endpoint + f"v2/teams/{team_slug}/items/stage",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id)],
                            "dataset_ids": [dataset_id],
                        },
                        "stage_id": stage_id,
                        "workflow_id": workflow_id,
                    }
                )
            ],
            json={},
        )
        item.set_stage(stage_id, workflow_id)


def test_assign(item: Item) -> None:
    with responses.RequestsMock() as rsps:
        team_slug = item.meta_params["team_slug"]
        dataset_id = item.meta_params["dataset_id"]
        assignee_id = 123456
        workflow_id = "123456"
        rsps.add(
            rsps.POST,
            item.client.config.api_endpoint + f"v2/teams/{team_slug}/items/assign",
            status=200,
            match=[
                json_params_matcher(
                    {
                        "filters": {
                            "item_ids": [str(item.id)],
                            "dataset_ids": [dataset_id],
                        },
                        "assignee_id": assignee_id,
                        "workflow_id": workflow_id,
                    }
                )
            ],
            json={},
        )
        item.assign(assignee_id, workflow_id)



from pathlib import Path
from typing import Iterable

import darwin.datatypes as dt
from darwin.exporter.formats.helpers.yolo_class_builder import (
    ClassIndex,
    build_class_index,
    export_file,
    save_class_index,
)


def export(annotation_files: Iterable[dt.AnnotationFile], output_dir: Path) -> None:
    """
    Exports the given ``AnnotationFile``\\s into the YOLO format inside of the given
    ``output_dir``.

    Parameters
    ----------
    annotation_files : Iterable[dt.AnnotationFile]
        The ``AnnotationFile``\\s to be exported.
    output_dir : Path
        The folder where the new pascalvoc files will be.
    """

    annotation_files = list(annotation_files)

    class_index = build_class_index(annotation_files)

    for annotation_file in annotation_files:
        export_file(annotation_file, class_index, output_dir, _build_txt)

    save_class_index(class_index, output_dir)


def _build_txt(annotation_file: dt.AnnotationFile, class_index: ClassIndex) -> str:
    yolo_lines = []
    for annotation in annotation_file.annotations:
        annotation_type = annotation.annotation_class.annotation_type

        if isinstance(annotation, dt.VideoAnnotation):
            raise ValueError(
                "YOLO format does not support video annotations for export or conversion."
            )

        if annotation_type == "bounding_box":
            data = annotation.data
        elif annotation_type == "polygon":
            data = annotation.data
            data = data.get("bounding_box")
        else:
            continue

        if annotation.data is None:
            continue
        if annotation_file.image_height is None or annotation_file.image_width is None:
            continue

        i = class_index[annotation.annotation_class.name]
        # x, y should be the center of the box
        # x, y, w, h are normalized to the image size
        x = data["x"] + data["w"] / 2
        y = data["y"] + data["h"] / 2
        w = data["w"]
        h = data["h"]
        imh = annotation_file.image_height
        imw = annotation_file.image_width
        x = x / imw
        y = y / imh
        w = w / imw
        h = h / imh

        yolo_lines.append(f"{i} {x} {y} {w} {h}")
    return "\n".join(yolo_lines)



from darwin.future.core.datasets.create_dataset import create_dataset
from darwin.future.core.datasets.get_dataset import get_dataset
from darwin.future.core.datasets.list_datasets import list_datasets
from darwin.future.core.datasets.remove_dataset import remove_dataset
